from importlib.resources import files

from pandas import DataFrame, read_csv, Series
import logging

from .base import EnergyUseSimulation
from .occupancy import OccupationSimulation
from ..types import OpenBESSpecification, LIGHTING_TECHNOLOGIES, LIGHTING_BALLASTS, LIGHTING_CONTROL, ENERGY_SOURCES

logger = logging.getLogger(__name__)


class LightingSimulation(EnergyUseSimulation):
    occupancy: OccupationSimulation
    _lighting_heat: float
    
    def __init__(self, spec: OpenBESSpecification, occupancy: OccupationSimulation = None):
        super().__init__(spec=spec)
        self.occupancy = occupancy or OccupationSimulation(spec=spec)

    def get_w_per_luminaire(self, zone: int) -> float:
        """kWh per day for a specific zone based on lighting system specifications.
        """
        lamp_number = getattr(self.spec, f"lighting_system_lamp_number_z{zone}")
        tech = getattr(self.spec, f"lighting_system_tech_z{zone}")
        ballast = getattr(self.spec, f"lighting_system_ballast_z{zone}")
        lamp_power = getattr(self.spec, f"lighting_system_lamp_power_z{zone}")
    
        try:
            if tech in [
                LIGHTING_TECHNOLOGIES.IC,
                LIGHTING_TECHNOLOGIES.HAL,
                LIGHTING_TECHNOLOGIES.LED
            ]:
                return float(lamp_power * lamp_number)
            if ballast == LIGHTING_BALLASTS.BE and tech == LIGHTING_TECHNOLOGIES.FT_T8:
                d = read_csv(str(files('openbes.simulations.lighting_data') / "lamp_ft_t8_be.csv"))
                return float(d.loc[(d["lamp_power"] == lamp_power)][f"lamp_number_{lamp_number}"].values[0])
            d = read_csv(str(files("openbes.simulations.lighting_data") / f"lamp_{tech.name.lower()}.csv"))
            return float(d.loc[(d["lamp_power"] == lamp_power)][f"lamp_number_{lamp_number}"].values[0])
        except (AttributeError, FileNotFoundError, KeyError, IndexError) as e:
            logger.warning(f"Badly matched self.spec for zone {zone} [{e.__class__.__name__}: {e}]")
        except Exception as e:
            logger.error(e, exc_info=True)
    
        return 0.0
    
    def get_kwh_per_day_for_zone(self, zone: int) -> float:
        """Calculate the kWh per day for a specific zone based on lighting system specifications.
        Args:
            self.spec (OpenBESSpecification): The building specifications self.spec data class.
            zone (int): The zone number to calculate kWh/day for.
        Returns:
            float: The kWh per day for the specified zone.
        """
        power_per_luminaire = self.get_w_per_luminaire(zone)
        luminary_number = getattr(self.spec, f"lighting_system_luminary_number_z{zone}")
        if luminary_number is None:
            logger.warning("Inadequate data to calculate lighting power for zone %d", zone)
            return 0.0
    
        power_per_zone = power_per_luminaire * luminary_number
    
        try:
            zone_number = getattr(self.spec, f"lighting_system_similar_zone_number_z{zone}")
            operating_hours = getattr(self.spec, f"lighting_system_operating_hours_z{zone}")
            simultaneity_factor = getattr(self.spec, f"lighting_system_simultaneity_factor_z{zone}")
    
            return power_per_zone * zone_number * simultaneity_factor * operating_hours / 1000.0
        except AttributeError:
            logger.warning("Inadequate data to calculate lighting power for zone %d", zone)
        except Exception as e:
            logger.error(e, exc_info=True)
        return 0.0
    
    def get_kwh_per_day_per_zone(self) -> DataFrame:
        """Calculate the kWh per day per zone based on lighting system specifications.
        Args:
            self.spec (OpenBESSpecification): The building specifications self.spec data class.
        Returns:
            DataFrame: A DataFrame containing kWh per day per zone.
        """
        zones = [f"Zone type {i}" for i in range(1, 7)]
        return DataFrame(
            [self.get_kwh_per_day_for_zone(i) for i in range(1, 7)],
            index=zones,
            columns=["kWh/day"]
        )
    
    def get_kwh_per_month_per_zone(self) -> DataFrame:
        """
        Calculate the kWh per month per zone based on lighting system specifications and operational hours per month.
        Args:
            self.spec (OpenBESSpecification): The building specifications self.spec data class.
        Returns:
            DataFrame: kWh used by each zone in each requested month
        """
        df = self.get_kwh_per_day_per_zone()
        operational_days_df = self.occupancy.occupied_days_per_month.copy()
        cross = df["kWh/day"].values[:, None] * operational_days_df.values
        return DataFrame(cross, columns=operational_days_df.index, index=df.index)
    
    def get_kwh_per_month(self) -> Series:
        """
        Calculate the kWh used by lighting in each month.
        Args:
            self.spec (OpenBESSpecification): The building specifications self.spec data class.
        Returns:
            DataFrame: The energy used on lighting in the building in kWh.
        """
        per_month = self.get_kwh_per_month_per_zone().sum()
        per_month.name = "kWh/month"
        return per_month

    @property
    def lights_on(self) -> 'Series[bool]':
        """Hourly lights on status.
        [Hourly Simulation column KJ]
    
        Lights are considered to be on during occupied hours between the specified on and off times.
        """
        if 'lights_on' not in self._hours.columns:
            df = self.occupancy.occupancy
            self._hours['lights_on'] = False
            on_time = self.spec.lighting_on_time
            off_time = self.spec.lighting_off_time
            if on_time is not None and off_time is not None:
                i = df.index.names.index('hour')
                self._hours['lights_on'] = df.apply(
                    lambda r: r['is_occupied'] and (on_time < r.name[i] <= off_time), axis=1
                )
        return self._hours['lights_on']

    @property
    def lighting_ratio(self) -> 'Series[float]':
        """Calculate the lighting ratio based on building specifications.
        [Hourly Simulation column KI]
        """
        if 'lighting_ratio' not in self._hours.columns:
            self._hours['lighting_ratio'] = self.lights_on * self.spec.lighting_simultaneity_factor
        return self._hours['lighting_ratio']

    @property
    def parasitic_heat(self) -> float:
        """Calculate the parasitic heat from lighting in W/m2.
        W,pc [Hourly Simulation column KR]
    
        Parasitic heat from lighting is modelled using EN 15193 Annex F which gives typical yearly values
        for school buildings.
        """
        if self.spec.parameters.lighting_on_off is not None and not self.spec.parameters.lighting_on_off:
            return 0.0
        # (emergency_kWh_per_m2_year + standby_kWh_per_m2_year) / hours_per_year * 1000 kW_per_W
        return (1 + 5) / 8760 * 1000

    @property
    def lighting_heat(self) -> float:
        """Calculate the lighting heat gains in W/m2.
        [Inputs cell C147]
    
        Heat from lighting is modelled using EN 15193 Annex F which gives typical yearly values
        for school buildings.
    
        The model accounts for differences in automatic and manual lighting controls.
        """
        if not hasattr(self, '_lighting_heat') or self._lighting_heat is None:
            if self.spec.parameters.lighting_on_off is not None and not self.spec.parameters.lighting_on_off:
                return 0.0
            FC = 1.0  # correction factor for building type (schools)
            kWh_per_year_m2 = 20.0
            tD = 1800.0  # daylight usage hours
            tN = 200.0  # night usage hours
            FD = 1.0 if self.spec.lighting_control == LIGHTING_CONTROL.Manual else 0.9  # daylight control factor
            F0 = 1.0 if self.spec.lighting_control == LIGHTING_CONTROL.Manual else 0.9  # general control factor
            self._lighting_heat = (
                    (FC * kWh_per_year_m2 / 1000 * ((tD * FD * F0) + (tN * F0))) *
                    (5 / 8760 * (8760 - (tD + tN))) *
                    1000 /
                    8760
            )
        return self._lighting_heat

    @property
    def energy_use(self) -> DataFrame:
        """Lighting energy use in kWh for each hour of the year.

        NOTE: This currently averages annual energy use by lighting over each hour of the year.
         A more sensible implementation will spread the usage over occupied hours only,
         taking into account lighting on/off times and lighting ratio.
        """
        if self._energy_use[ENERGY_SOURCES.Electricity].hasnans:
            self._energy_use[ENERGY_SOURCES.Electricity] = self.get_kwh_per_month().sum().sum() / len(self._energy_use)
        return self._energy_use
        



import pvlib
import numpy as np
from pandas import DataFrame, Series, DatetimeIndex

from .base import HOURS_DF
from ..types import COMPASS_POINTS


class SolarIrradiationSimulation:
    """
    Simulation class for solar irradiation data from EPW files.

    Uses pvlib but may have to fall back to copying Excel calculations.
    """
    _hours: DataFrame
    location: pvlib.location.Location
    times: DatetimeIndex
    _solarposition: DataFrame
    _solar_irradiation: DataFrame
    _solar_declination_: np.array = None
    _hour_angle_: np.array = None

    def __init__(self, epw_data: DataFrame, epw_metadata: dict):
        self._hours = HOURS_DF.copy()
        self.epw_data = epw_data
        self.epw_metadata = epw_metadata
        tz = epw_metadata['TZ']
        self.location = pvlib.location.Location(
            latitude=epw_metadata['latitude'],
            longitude=epw_metadata['longitude'],
            tz=tz,
            altitude=epw_metadata['altitude']
        )
        self._solar_irradiation = DataFrame()
        self._solar_irradiation.index = self._hours.index

    @property
    def lat(self) -> float:
        """Latitude of the location from EPW metadata."""
        return round(self.epw_metadata.get('latitude'), 3)

    @property
    def lon(self) -> float:
        """Longitude of the location from EPW metadata."""
        return round(self.epw_metadata.get('longitude'), 2)

    @property
    def timezone(self) -> float:
        """Timezone of the location from EPW metadata."""
        return self.epw_metadata.get('TZ')

    @property
    def altitude(self) -> float:
        """Altitude of the location from EPW metadata."""
        return self.epw_metadata.get('altitude')

    @property
    def ghi(self) -> 'Series[float]':
        """Global Horizontal Irradiance (GHI) from EPW data."""
        if 'global_horizontal_irradiance' not in self._hours.columns:
            self._hours['global_horizontal_irradiance'] = list(self.epw_data['ghi'].astype(float))
        return self._hours['global_horizontal_irradiance']

    @property
    def dni(self) -> 'Series[float]':
        """Direct Normal Irradiance (DNI) from EPW data."""
        if 'direct_normal_irradiance' not in self._hours.columns:
            self._hours['direct_normal_irradiance'] = list(self.epw_data['dni'].astype(float))
        return self._hours['direct_normal_irradiance']

    @property
    def dhi(self) -> 'Series[float]':
        """Diffuse Horizontal Irradiance (DHI) from EPW data."""
        if 'diffuse_horizontal_irradiance' not in self._hours.columns:
            self._hours['diffuse_horizontal_irradiance'] = list(self.epw_data['dhi'].astype(float))
        return self._hours['diffuse_horizontal_irradiance']

    @property
    def _hour_angle(self):
        """Hour angle (h) in radians for each hour."""
        if self._hour_angle_ is None:
            local_standard_time = np.array(
                self._hours.index.get_level_values(self._hours.index.names.index('hour'))
            )
            day_of_year = np.array(self._hours.index.get_level_values(self._hours.index.names.index('day')))
            orbital_position = (
                    2 * np.pi *
                    (day_of_year - 1) /
                    365
            )
            equation_of_time = 2.2918 * (
                    0.0075 +
                    0.1868 * np.cos(orbital_position) -
                    3.2077 * np.sin(orbital_position) -
                    1.4615 * np.cos(2 * orbital_position) -
                    4.089 * np.sin(2 * orbital_position)
            )
            longitude_of_local_meridian = 15 * self.timezone
            apparent_solar_time = (
                    local_standard_time +
                    equation_of_time / 60 +
                    (self.lon - longitude_of_local_meridian) / 15
            )
            self._hour_angle_ = np.radians(15 * (apparent_solar_time - 12))
        return self._hour_angle_

    @property
    def _solar_declination(self):
        """Solar declination (delta) in radians for each hour."""
        if self._solar_declination_ is None:
            day_of_year = np.array(self._hours.index.get_level_values(self._hours.index.names.index('day')))
            self._solar_declination_ = np.radians(23.45 * np.sin(2 * np.pi * (284 + day_of_year) / 365))
        return self._solar_declination_

    @property
    def solar_altitude(self) -> 'Series[float]':
        """Solar altitude angle (beta) in degrees for each hour.
        [Solar radiation column P]
        """
        if 'solar_altitude' not in self._hours.columns:
            latitude = np.radians(self.lat)
            sin_solar_altitude = (
                np.cos(latitude) * np.cos(self._solar_declination) * np.cos(self._hour_angle) +
                np.sin(latitude) * np.sin(self._solar_declination)
            )
            self._hours['solar_altitude'] = np.degrees(np.asin(sin_solar_altitude))
        return self._hours['solar_altitude']

    @property
    def solar_zenith(self) -> 'Series[float]':
        """Solar zenith angle (theta) in degrees for each hour.
        """
        return 90.0 - self.solar_altitude

    @property
    def solar_azimuth(self) -> 'Series[float]':
        """Solar azimuth angle (phi) in degrees for each hour.
        [Solar radiation column V]
        """
        if 'solar_azimuth_degrees' not in self._hours.columns:
            altitude_rad = np.radians(self.solar_altitude)
            lat_rad = np.radians(self.lat)
            sin_phi = (
                    np.sin(self._hour_angle) *
                    np.cos(self._solar_declination)
            ) / np.cos(altitude_rad)
            cos_phi = (
                np.cos(self._hour_angle) * np.cos(self._solar_declination) * np.sin(lat_rad) -
                np.sin(self._solar_declination) * np.cos(lat_rad)
            ) / np.cos(altitude_rad)
            phi_from_sin = np.degrees(np.asin(sin_phi))
            phi_from_cos = np.degrees(np.acos(cos_phi))
            from_cos = sin_phi > 0
            from_sin = cos_phi > 0
            others = ~(from_cos | from_sin)
            self._hours['solar_azimuth_degrees'] = (
                phi_from_cos * from_cos +
                phi_from_sin * (from_sin & ~from_cos) +
                (-180 - phi_from_sin) * others
            )
        return self._hours['solar_azimuth_degrees']

    def get_solar_irradiation(self, compass_point: COMPASS_POINTS) -> 'Series[float]':
        """Get the hourly solar irradiation on a vertical surface facing the given compass point in Wh/m2.
        [Hourly simulation columns M:T, Solar radiation BT:CA]
        """
        if compass_point not in self._solar_irradiation.columns:
            # These values in Hourly simulation M114:T114 are used in the map, then adjusted to PVLib convention
            surface_azimuth = {
                COMPASS_POINTS.North: 180,
                COMPASS_POINTS.NorthEast: 360 - 135,
                COMPASS_POINTS.East: 360 - 90,
                COMPASS_POINTS.SouthEast: 360 - 45,
                COMPASS_POINTS.South: 0,
                COMPASS_POINTS.SouthWest: 45,
                COMPASS_POINTS.West: 90,
                COMPASS_POINTS.NorthWest: 135,
            }[compass_point]
            solar_azimuth = self.solar_azimuth
            gamma = abs(solar_azimuth - surface_azimuth)  # [Solar radiation columns W:AD]
            solar_altitude_rad = np.radians(self.solar_altitude)
            aoi = np.cos(solar_altitude_rad) * np.cos(np.radians(gamma))  # [Solar radiation columns AE:AL]
            aoi = np.degrees(np.arccos(aoi))  # [Solar radiation columns AM:AT]
            # [Solar radiation columns AU:BB]
            beam_component = np.logical_not((90 < gamma) & (gamma < 270)) * (aoi >= 0) * (self.dni * np.cos(np.radians(aoi)))
            # [Solar radiation columns BC:BJ]
            diffuse_component_ratio = (
                0.55 +
                0.437 * np.cos(np.radians(aoi)) +
                0.313 * (np.cos(np.radians(aoi)) ** 2)
            )
            # [Solar radiation columns BK:BR; capping is actually done in the previous column set]
            diffuse_component = np.maximum(0.45, diffuse_component_ratio) * self.dhi
            # [Solar radiation column BS]
            ground_component = (self.dni * np.sin(solar_altitude_rad) + self.dhi) * 0.14 / 2
            self._solar_irradiation[compass_point] = beam_component + diffuse_component + ground_component
        return self._solar_irradiation[compass_point]

    @property
    def solar_irradiation(self) -> DataFrame:
        """Hourly solar irradiation on a horizontal surface in Wh/m2, columns are COMPASS_POINTS.
        [Hourly simulation columns M:T, Solar radiation BT:CA]
        """
        if self._solar_irradiation.empty:
            for compass_point in list(COMPASS_POINTS):
                self.get_solar_irradiation(compass_point)
        return self._solar_irradiation


from pandas import DataFrame

from ..types import OpenBESSpecification, ENERGY_SOURCES


def month_for_day(day_number_in_year: int) -> int:
    """Calculate the month for a given day number in the year.
    Args:
        day_number_in_year (int): The day number in the year (1-365).
    Returns:
        int: The corresponding month (1-12).
    """
    if day_number_in_year <= 31:
        return 1
    elif day_number_in_year <= 59:
        return 2
    elif day_number_in_year <= 90:
        return 3
    elif day_number_in_year <= 120:
        return 4
    elif day_number_in_year <= 151:
        return 5
    elif day_number_in_year <= 181:
        return 6
    elif day_number_in_year <= 212:
        return 7
    elif day_number_in_year <= 243:
        return 8
    elif day_number_in_year <= 273:
        return 9
    elif day_number_in_year <= 304:
        return 10
    elif day_number_in_year <= 334:
        return 11
    else:
        return 12


# Blank DataFrame of each hour with month info, indexed by day of the year
HOURS_DF = DataFrame([
    {
        'month': month_for_day(d),
        'day': d,
        'hour': h,
        'is_daytime': 8 <= h <= 22
    } for d in range(1, 366) for h in range(1, 25)
]).set_index(['month', 'day', 'hour'])


class HourlySimulation:
    """
    Base class for hourly simulations.

    Each instance is initialized with an OpenBESSpecification and contains an _hours property
    that holds a DataFrame representing each hour of the year.

    Properties will typically add columns to the _hours DataFrame as needed for various calculations,
    and usually return that DataFrame or specific columns from it.
    """

    spec: OpenBESSpecification
    _hours: DataFrame

    def __init__(self, spec: OpenBESSpecification):
        self.spec = spec
        self._hours = HOURS_DF.copy()


class EnergyUseSimulation(HourlySimulation):
    """
    Base class for hourly energy use simulations.

    Child classes should override the energy_use property to return a DataFrame representing
    the energy use in kW for each hour of the year, with a column for each of the ENERGY_SOURCES.

    The _energy_use property is initialized as a DataFrame of NaNs with the same index as _hours
    and columns for each ENERGY_SOURCES. self.energy_use should populate this DataFrame appropriately.
    """
    def __init__(self, spec: OpenBESSpecification):
        super().__init__(spec)
        self._energy_use = DataFrame(index=self._hours.index, columns=list(ENERGY_SOURCES)).astype(float)

    @property
    def energy_use(self) -> DataFrame:
        """DataFrame of energy use in kW for each hour of the year, with a column for each ENERGY_SOURCES.
        """
        raise NotImplementedError("Child classes must implement the energy_use property.")

    @property
    def annual_energy_use(self) -> DataFrame:
        return self.energy_use.sum().sum()



import logging
from math import atan
from typing import List

import numpy as np
from pandas import Series

from .base import EnergyUseSimulation
from .climate import ClimateSimulation, RELATIVE_HUMIDITY
from .geometry import BuildingGeometry
from .lighting import LightingSimulation
from .occupancy import OccupationSimulation
from .ventilation import VentilationSimulation
from ..types import OpenBESSpecification, COOLING_SYSTEM_TYPES

logger = logging.getLogger(__name__)

MIN_COOLING_CAPACITY = 0.01  # kW
MIN_COOLING_EFFICIENCY = 0.01  # kWh


class CoolingSystemSimulation(EnergyUseSimulation):
    """A class to simulate a cooling system's energy consumption.

    [Cell references are for System 1]
    """
    system_number: int
    _nominal_consumption: float = None
    _nominal_capacity: float = None
    _sensible_nominal_capacity: float = None
    _efficiency: float = None
    _area: float = None
    _Ts_int: float = None
    _Th_int: float = None

    def __init__(
            self,
            spec: OpenBESSpecification,
            system_number: int = 1,
            climate: ClimateSimulation = None,
    ):
        super().__init__(spec)
        self.system_number = system_number
        self.climate = climate or ClimateSimulation(spec)
        self.geometry = climate.geometry
        self.occupancy = climate.occupancy

    def _attr(self, attr_name: str):
        return getattr(self.spec, f"cooling_system{self.system_number}_{attr_name}")

    @property
    def area(self) -> float:
        """Total area affected by cooling system in m2."""
        if self._area is None:
            areas = self.geometry.conditioned_floor_areas.groupby('zone').sum()
            simultaneity = [self._attr(f"simultaneity_factor_{z.value.split('_')[0]}") for z in areas.index]
            self._area = (areas * simultaneity).sum()
        return self._area

    @property
    def ratio(self) -> 'Series[float]':
        """Hourly proportion of total cooling capacity that is enabled.

        The system will typically be 100% enabled during occupied hours.
        """
        return self.occupancy.occupancy['is_occupied'].astype(float)

    def _get_target_temp(self) -> 'np.array[float]':
        """Get the target temperature for cooling based on thresholds and air temperature."""
        summer_temp = np.array(self.climate.set_point_temperature['min_temp_set_point'])
        winter_temp = np.array(self.climate.set_point_temperature['max_temp_set_point'])
        return np.where(
            self.climate.air_free_temp < winter_temp,
            winter_temp,
            np.where(
                self.climate.air_free_temp > summer_temp,
                summer_temp,
                self.climate.air_free_temp
            )
        )

    @property
    def phi_hc_nd_actual(self) -> 'Series[float]':
        """Cooling load per unit area in W/m2.
        [Hourly simulation column GW]

        The system is on when
        - the system is enabled (ratio > 0.0)
        - temperature > target temperature

        When on, the system operates at the capacity required to bring the temperature within tolerance.
        """
        if 'phi_hc_nd_actual' not in self._hours.columns:
            target_temp = self._get_target_temp()
            self._hours['phi_hc_nd_actual'] = (
                10 *
                (target_temp - self.climate.air_free_temp_hc_0) /
                (self.climate.air_free_temp_hc_10 - self.climate.air_free_temp_hc_0)
            ) * self.ratio * (target_temp < self.climate.air_free_temp)
        return self._hours['phi_hc_nd_actual']

    @property
    def phi_c_nd_ac(self) -> float:
        """Cooling load per unit area in W/m2.
        [Hourly simulation column GY]
        """
        if 'phi_c_nd_ac' not in self._hours.columns:
            self._hours['phi_c_nd_ac'] = self.phi_hc_nd_actual.apply(
                lambda r: min(r, 0.0) if r < (
                        getattr(self.spec.parameters, f"cooling_system{self.system_number}_min_demand") * -1
                ) else 0.0
            ) * self.spec.parameters.cooling_load_factor
        return self._hours['phi_c_nd_ac']

    @property
    def demand(self) -> 'Series[float]':
        """Hourly cooling demand in kW.
        [Hourly simulation column HA]
        """
        if 'demand' not in self._hours.columns:
            self._hours['demand'] = - (
                self.phi_c_nd_ac * self.area
            ) / 1000  # W -> kW
        return self._hours['demand']

    @property
    def Ts_int(self) -> float:
        """Internal set point temperature. ???????
        [HD102]
        """
        if self._Ts_int is None:
            self._Ts_int = self.spec.setpoint_summer_day - self.spec.parameters.temperature_tolerance
        return self._Ts_int

    @property
    def Th_int(self) -> float:
        """Internal heat temperature. ???????
        [HD101]
        """
        if self._Th_int is None:
            self._Th_int = (
                    self.Ts_int * atan(0.151977 * (RELATIVE_HUMIDITY + 8.313659)**0.5) +
                    atan(self.Ts_int + RELATIVE_HUMIDITY) -
                    atan(RELATIVE_HUMIDITY - 1.676331) +
                    0.00391838 * RELATIVE_HUMIDITY**1.5 * atan(0.023101 * RELATIVE_HUMIDITY) -
                    4.686035
            )
        return self._Th_int

    @property
    def fan_cooling_power(self) -> 'Series[float]':
        """Hourly fan cooling power (reference value).
        [Hourly simulation column HF, HN]
        """
        if 'fan_cooling_power' not in self._hours.columns:
            self._hours['fan_cooling_power'] = self.demand / self.cap_sen_ref
        return self._hours['fan_cooling_power']

    @property
    def cap_sen_ref(self) -> 'Series[float]':
        """Reference sensible cooling capacity. ???????
        [Hourly simulation column HH]
        """
        if 'cap_sen_ref' not in self._hours.columns:
            return self.sensible_nominal_capacity * self.cap_ref_t
        return self._hours['cap_sen_ref']

    @property
    def cap_ref_t(self) -> 'Series[float]':
        """Reference cooling capacity given the temperature difference. ???????
        [Hourly simulation column HI]
        """
        if 'cap_ref_t' not in self._hours.columns:
            self._hours['cap_ref_t'] = (
               0.500601825 -
               0.046438331 * self.Th_int -
               0.000324724 * (self.Th_int**2) +
               0.069957819 * self.Ts_int -
               0.0000342756 * (self.Ts_int**2) -
               0.013202081 * self.climate.dry_bulb_temp +
               0.0000793065 * (self.climate.dry_bulb_temp**2)
            )
        return self._hours['cap_ref_t']

    @property
    def con_ref_t(self) -> 'Series[float]':
        """Reference consumption given the temperature difference. ???????
        [Hourly simulation column HL]
        """
        if 'con_ref_t' not in self._hours.columns:
            self._hours['con_ref_t'] = (
               0.1117801 +
               0.028493334 * self.Th_int -
               0.000411156 * (self.Th_int**2) +
               0.021414276 * self.climate.dry_bulb_temp + 0.000161125 * (self.climate.dry_bulb_temp**2) -
               0.000679104 * self.climate.dry_bulb_temp * self.Th_int
            )
        return self._hours['con_ref_t']

    @property
    def con_ref_fcp(self) -> 'Series[float]':
        """Reference consumption fan cooling power. ????????
        [Hourly simulation column HM]
        """
        if 'con_ref_fcp' not in self._hours.columns:
            self._hours["con_ref_fcp"] = (
                0.2012307 -
                0.0312175 * self.fan_cooling_power +
                1.9504979 * (self.fan_cooling_power**2) -
                1.1205104 * (self.fan_cooling_power**3)
            )
        return self._hours['con_ref_fcp']

    @property
    def number(self) -> int:
        """Number of systems of this type installed.
        """
        return self._attr('number')

    @property
    def nominal_capacity(self) -> float:
        """
        """
        if self._nominal_capacity is None:
            self._nominal_capacity = max(self._attr('nominal_capacity'), MIN_COOLING_CAPACITY) * self.number
        return self._nominal_capacity

    @property
    def sensible_nominal_capacity(self) -> float:
        """Cooling system nominal sensible refrigeration capacity in kW.
        [Hourly simulation cell HD91]
        """
        if self._sensible_nominal_capacity is None:
            self._sensible_nominal_capacity = max(
                self._attr('sensible_nominal_capacity'),
                MIN_COOLING_CAPACITY
            ) * self.number
        return self._sensible_nominal_capacity

    @property
    def efficiency(self) -> float:
        """
        """
        if self._efficiency is None:
            # NB: Typo in 'energy_efficifiency_ratio' is in the original spec
            self._efficiency = max(self._attr('energy_efficifiency_ratio'), MIN_COOLING_EFFICIENCY)
        return self._efficiency
    
    @property
    def nominal_consumption(self) -> float:
        """Cooling system nominal refrigeration consumption in kWh.
        [Hourly simulation cell HD92]
        """
        if self._nominal_consumption is None:
            self._nominal_consumption = self.nominal_capacity / self.efficiency
        return self._nominal_consumption


    @property
    def energy_use(self) -> 'Series[float]':
        """Cooling system energy use in kWh for each hour of the year for each ENERGY_SOURCES.
        [Hourly outputs column Q, disaggregated; Hourly simulation column HK]
        """
        if self._energy_use[self._attr('energy_source')].hasnans:
            self._energy_use = self._energy_use.fillna(0.0).infer_objects(copy=False)
            if self._attr('type') == COOLING_SYSTEM_TYPES.Heat_pump:
                self._energy_use[self._attr('energy_source')] = (
                        (self.demand != 0.0) * self.con_ref_t * self.con_ref_fcp * self.nominal_consumption
                )
        return self._energy_use


class CoolingSimulation(EnergyUseSimulation):
    """A class to simulate cooling energy consumption based on building specifications."""
    cooling_simulations: List[CoolingSystemSimulation]

    def __init__(
            self,
            spec: OpenBESSpecification,
            geometry: BuildingGeometry = None,
            occupancy: OccupationSimulation = None,
            lighting: LightingSimulation = None,
            ventilation: VentilationSimulation = None,
            climate: ClimateSimulation = None,
    ):
        super().__init__(spec)
        geometry = geometry or BuildingGeometry(self.spec)
        occupancy = occupancy or OccupationSimulation(self.spec, geometry=geometry)
        lighting = lighting or LightingSimulation(self.spec, occupancy=occupancy)
        ventilation = ventilation or VentilationSimulation(self.spec, occupancy=occupancy, geometry=geometry)
        self.climate_simulation = climate or ClimateSimulation(
            spec,
            geometry=geometry,
            occupancy=occupancy,
            lighting=lighting,
            ventilation=ventilation,
        )
        self.cooling_simulations = []
        while True:
            system_number = len(self.cooling_simulations) + 1
            attr_name = f"cooling_system{system_number}_type"
            if not hasattr(spec, attr_name):
                break
            self.cooling_simulations.append(
                CoolingSystemSimulation(
                    spec=spec,
                    system_number=system_number,
                    climate=self.climate_simulation
                )
            )
    
    @property
    def energy_use(self) -> 'Series[float]':
        """Cooling energy use in kWh for each hour of the year for each ENERGY_SOURCES.
        [Hourly outputs column Q], disaggregated
        """
        return sum([x.energy_use for x in self.cooling_simulations])



"""
Helper functions to simulate occupancy patterns in buildings.
"""
import logging

from pandas import DataFrame, Series

from .base import HourlySimulation
from .geometry import BuildingGeometry
from ..types import (
    DAYS,
    OpenBESSpecification,
    OCCUPATION_ZONES,
    FLOORS,
    MONTHS,
)

logger = logging.getLogger(__name__)

# Occupation m2 per person for different zones (CTE DB-SI Table 2.1, Database cells R42:S46)
M2_PER_PERSON = DataFrame([
    {"zone": OCCUPATION_ZONES.Office, "m2_per_person": 5},
    {"zone": OCCUPATION_ZONES.Teaching, "m2_per_person": 1.5},
    {"zone": OCCUPATION_ZONES.Canteen, "m2_per_person": 5},
    {"zone": OCCUPATION_ZONES.Common_areas, "m2_per_person": 5},
    {"zone": OCCUPATION_ZONES.Other, "m2_per_person": 5},
]).set_index("zone")


def day_of_the_week(day_number_in_year: int) -> DAYS:
    """Calculate the day of the week for a given day number in the year.
    Args:
        day_number_in_year (int): The day number in the year (1-365).
    Returns:
        DAYS: The corresponding day of the week.
    """
    return DAYS.get_by_index((day_number_in_year - 1) % 7)


def is_public_holiday(day_number_in_year: int) -> bool:
    """Check if a given day number in the year is a public holiday.
    Args:
        day_number_in_year (int): The day number in the year (1-365).
    Returns:
        bool: True if the day is a public holiday, False otherwise.
    """
    # Example public holidays (day numbers in the year)
    if day_number_in_year <= 5:
        return True  # First week of January
    return day_number_in_year >= 358  # Every day after Xmas is a holiday


class OccupationSimulation(HourlySimulation):
    geometry: BuildingGeometry
    _occupation_ratio: float
    _occupation_m2_per_person: float
    _metabolic_rate_per_m2: float
    _occupied_days_per_month: Series

    def __init__(self, spec: OpenBESSpecification, geometry: BuildingGeometry = None):
        super().__init__(spec)
        self.geometry = geometry or BuildingGeometry(spec=self.spec)

    def is_occupied_month(self, month: int) -> bool:
        """Determine if a given month is an occupied month.
        """
        if month == 1:
            return self.spec.schedule_january
        if month == 2:
            return self.spec.schedule_february
        if month == 3:
            return self.spec.schedule_march
        if month == 4:
            return self.spec.schedule_april
        if month == 5:
            return self.spec.schedule_may
        if month == 6:
            return self.spec.schedule_june
        if month == 7:
            return self.spec.schedule_july
        if month == 8:
            return self.spec.schedule_august
        if month == 9:
            return self.spec.schedule_september
        if month == 10:
            return self.spec.schedule_october
        if month == 11:
            return self.spec.schedule_november
        if month == 12:
            return self.spec.schedule_december
        raise ValueError("Invalid month")

    def is_occupied_day(self, day_number_in_year: int) -> bool:
        """Determine if a given day number in the year is an occupied day.
        Args:
            day_number_in_year (int): The day number in the year (1-365).
        Returns:
            bool: True if the day is occupied, False otherwise.
        """
        if not self.spec.holiday and is_public_holiday(day_number_in_year):
            return False
        day = day_of_the_week(day_number_in_year)
        if day == DAYS.Mon:
            return self.spec.schedule_monday
        if day == DAYS.Tue:
            return self.spec.schedule_tuesday
        if day == DAYS.Wed:
            return self.spec.schedule_wednesday
        if day == DAYS.Thu:
            return self.spec.schedule_thursday
        if day == DAYS.Fri:
            return self.spec.schedule_friday
        if day == DAYS.Sat:
            return self.spec.schedule_saturday
        if day == DAYS.Sun:
            return self.spec.schedule_sunday
        raise ValueError("Invalid day number in year")

    def get_zone_total_area(self, zone: OCCUPATION_ZONES, floor: FLOORS = None) -> float:
        """Get the total area for a given occupation zone.
        Args:
            zone (OCCUPATION_ZONES): The occupation zone.
            floor: Floor to consider. If None, all floors are considered.
        Returns:
            float: The total area of the zone in m².
        """
        if not hasattr(self, 'geometry') or self.geometry is None:
            self.geometry = BuildingGeometry(spec=self.spec)
        if floor is None:
            return self.geometry.gross_floor_areas.groupby(level='zone').sum().loc[zone]
        return self.geometry.gross_floor_areas.loc[(floor, zone)].sum()

    @property
    def occupation_ratio(self) -> float:
        """Calculate the occupation ratio (occupation/capacity) based on the building schedule.

        Returns:
            float: The occupation ratio (0.0 to 1.0).
        """
        if not hasattr(self, '_occupation_ratio') or self._occupation_ratio is None:
            capacity = self.spec.max_building_occupation
            current_occupation = self.spec.typical_occupation
            if current_occupation is None or current_occupation < 0:
                logger.warning(
                    "Cannot calculate occupation percentage without `typical_occupation`. Defaulting to 100% occupation."
                )
                return 1.0
            try:
                if capacity > 0 and current_occupation > 0:
                    return current_occupation / capacity
            except (ZeroDivisionError, TypeError):
                pass
            zonal_occupation_capacity = [
                self.get_zone_total_area(zone=z) / M2_PER_PERSON.loc[z, "m2_per_person"]
                for z in OCCUPATION_ZONES
            ]
            self._occupation_ratio = current_occupation / sum(zonal_occupation_capacity)
        return self._occupation_ratio

    @property
    def occupancy(self) -> DataFrame:
        """Generate an occupancy schedule by hour for the entire year.
        [Hourly Simulation column KH]

        DataFrame: with is_occupied and occupancy_ratio (0.0-1.0) for each hour of the year.
        """
        if 'is_occupied' not in self._hours.columns or 'occupancy_ratio' not in self._hours.columns:
            open_times = [self.spec.occupancy_open_office, self.spec.occupancy_open_canteen, self.spec.occupancy_open_teaching]
            close_times = [self.spec.occupancy_close_office, self.spec.occupancy_close_canteen, self.spec.occupancy_close_teaching]
            if all(ot is None for ot in open_times) or all(ct is None for ct in close_times):
                raise ValueError("Occupancy open and close times must be self.specified in the building specification.")
            open_time = min(ot for ot in open_times if ot is not None)
            close_time = max(ct for ct in close_times if ct is not None)

            self._hours['is_occupied'] = False
            self._hours['occupancy_ratio'] = 0.0
            # Get a mask for occupied hours in occupied days in occupied months that aren't public holidays
            index_df = self._hours.index.to_frame()
            month_mask = index_df['month'].apply(lambda m: self.is_occupied_month(m))
            day_mask = index_df['day'].apply(lambda d: self.is_occupied_day(d))
            hour_mask = (index_df['hour'] >= open_time) & (index_df['hour'] <= close_time)
            mask = month_mask & day_mask & hour_mask
            self._hours.loc[mask, 'is_occupied'] = True
            self._hours.loc[mask, 'occupancy_ratio'] = self.occupation_ratio

        return self._hours[['is_occupied', 'occupancy_ratio']]

    @property
    def occupied_days_per_month(self) -> Series:
        """Calculate the number of occupied days per month based on the building schedule.
        Returns:
            Series: A Series with the number of occupied days for each month.
        """
        if not hasattr(self, '_occupied_days_per_month') or self._occupied_days_per_month is None:
            self._occupied_days_per_month = (
                self.occupancy['is_occupied'].groupby(level=['day', 'month']).any().groupby(level='month').sum()
            )
            self._occupied_days_per_month.index = list(MONTHS)
        return self._occupied_days_per_month

    @property
    def occupation_m2_per_person(self) -> float:
        """Calculate the occupation area per person based on building specifications.
        m2/person [Inputs C139]

        The occupation m2/person is calculated as the building area for zones in which people typically inhabit
        (office and teaching) divided by the typical occupation (number of people).

        People are assumed to inhabit only office and teaching zones, because when they are in canteen zones,
        common zones, etc. they are still generating their metabolic heat load, just in a different location.
        Because locations are amalgamated in OpenBES, we can ignore _where_ people are, and just focus on
        how many people there are in total, and what area is available to them.

        Returns:
            float: The occupation area per person in m2/person.
        """
        if not hasattr(self, '_occupation_m2_per_person') or self._occupation_m2_per_person is None:
            occupied_zone_areas = (
                    self.get_zone_total_area(zone=OCCUPATION_ZONES.Office) +
                    self.get_zone_total_area(zone=OCCUPATION_ZONES.Teaching)
            ) * self.spec.parameters.nia_gba_ratio  # scale by net inhabitable area ratio
            office_population = (
                    self.get_zone_total_area(zone=OCCUPATION_ZONES.Office) /
                    M2_PER_PERSON.loc[OCCUPATION_ZONES.Office, "m2_per_person"] *
                    self.spec.parameters.nia_gba_ratio
            )
            teaching_population = (
                    self.get_zone_total_area(zone=OCCUPATION_ZONES.Teaching) /
                    M2_PER_PERSON.loc[OCCUPATION_ZONES.Teaching, "m2_per_person"] *
                    self.spec.parameters.nia_gba_ratio
            )
            simultaneity_factor = 0.75  # [Inputs cell F137]
            simultaneity_adjusted_population = (office_population + teaching_population) * simultaneity_factor
            self._occupation_m2_per_person = occupied_zone_areas / simultaneity_adjusted_population
        return self._occupation_m2_per_person

    @property
    def metabolic_rate_per_m2(self) -> float:
        """Calculate the metabolic rate per square meter based on building specifications.
        [Inputs cell C140, Database R32: Table G.10 ISO 13790]

        The more space each person has (m2/person), the lower the metabolic rate per square meter (W/m2).
        ISO 13790 provides typical values for metabolic rates based on occupation density.

        Args:
            self.spec (OpenBESSpecification): The building specifications self.spec data class.
        Returns:
            float: The metabolic rate per square meter in W/m2.
        """
        if not hasattr(self, '_metabolic_rate_per_m2') or self._metabolic_rate_per_m2 is None:
            if self.spec.parameters.occupancy_on_off is not None and not self.spec.parameters.occupancy_on_off:
                self._metabolic_rate_per_m2 = 0.0
            else:
                occupation_density = self.occupation_m2_per_person
                occupation_density_thresholds = [
                    {'threshold_m2_per_person': 1, 'metabolic_rate_W_per_m2': 15.0},
                    {'threshold_m2_per_person': 2, 'metabolic_rate_W_per_m2': 10.0},
                    {'threshold_m2_per_person': 5.5, 'metabolic_rate_W_per_m2': 5.0},
                    {'threshold_m2_per_person': 14, 'metabolic_rate_W_per_m2': 3.0},
                    {'threshold_m2_per_person': 20, 'metabolic_rate_W_per_m2': 2.0},
                ]
                for i in range(len(occupation_density_thresholds)):
                    if occupation_density < occupation_density_thresholds[i]['threshold_m2_per_person']:
                        self._metabolic_rate_per_m2 = occupation_density_thresholds[i]['metabolic_rate_W_per_m2']
                        break
            if not hasattr(self, '_metabolic_rate_per_m2') or self._metabolic_rate_per_m2 is None:
                logger.warning("Occupation density lower than 0.05 person/m2. Using minimum metabolic rate of 2 W/m2.")
                self._metabolic_rate_per_m2 = 2.0
        return self._metabolic_rate_per_m2



from typing import List

from pandas import Series
import logging
from pandas import DataFrame

from .base import EnergyUseSimulation
from .geometry import BuildingGeometry
from .occupancy import OccupationSimulation
from ..types import OpenBESSpecification

logger = logging.getLogger(__name__)

class VentilationSystemSimulation(EnergyUseSimulation):
    system_number: int
    geometry: BuildingGeometry
    _air_supply_rate_adjusted: float

    def __init__(
            self,
            spec: OpenBESSpecification,
            system_number: int = 1,
            occupancy: OccupationSimulation = None,
            geometry: BuildingGeometry = None
    ):
        super().__init__(spec=spec)
        self.system_number = system_number
        self.geometry = geometry or BuildingGeometry(spec=self.spec)
        self.occupancy = occupancy or OccupationSimulation(spec=spec)

    def _attr(self, attr_name: str):
        return getattr(self.spec, f"ventilation_system{self.system_number}_{attr_name}")

    @property
    def air_supply_rate_adjusted(self) -> float:
        """Air supply rate (m3/h/m2) adjusted for system efficiency.
        [Hourly simulation cells IV99, JB99]
        """
        if not hasattr(self, '_air_supply_rate_adjusted') or self._air_supply_rate_adjusted is None:
            rated_flow_rate = self._attr('airflow') / self._attr('ventilated_area')  # m3/h/m2
            efficiency = self._attr('heat_recovery_efficiency')
            if rated_flow_rate is None or efficiency is None:
                self._air_supply_rate_adjusted = 0.0
            else:
                self._air_supply_rate_adjusted = rated_flow_rate * (1 - efficiency)
        return self._air_supply_rate_adjusted

    @property
    def ventilation_on(self) -> 'Series[bool]':
        """Hourly ventilation status (on/off) throughout the year.
        [Hourly simulation columns IR, IX]

        Ventilation only runs between the specified on and off times,
        and only while the building is occupied.
        """
        if 'ventilation_on' not in self._hours.columns:
            on_time = self._attr('on_time')
            off_time = self._attr('off_time')
            self._hours['ventilation_on'] = list(
                map(lambda x: on_time <= x <= off_time, self._hours.index.get_level_values('hour').values)
            )
            self._hours['ventilation_on'] = self._hours['ventilation_on'] * self.occupancy.occupancy['is_occupied']
        return self._hours['ventilation_on']

    @property
    def air_supply_rate(self) -> 'Series[float]':
        """Hourly air supply rate (m3/h/m2) throughout the year.
        [Hourly simulation columns IV, JB]
        """
        if 'air_supply_rate' not in self._hours.columns:
            area = self.geometry.conditioned_floor_area
            rated_flow_rate = self.air_supply_rate_adjusted * self._attr('ventilated_area')
            if rated_flow_rate is None or area == 0:
                self._hours['air_supply_rate'] = 0.0
            else:
                self._hours['air_supply_rate'] = (
                        (rated_flow_rate / area) *
                        self.ventilation_on.astype(float)
                )
        return self._hours['air_supply_rate']

    @property
    def energy_use(self) -> DataFrame:
        """Ventilation energy use in kWh for each hour of the year for each ENERGY_SOURCE.
        """
        if self._energy_use[self._attr('energy_source')].hasnans:
            self._energy_use[self._attr('energy_source')] = (
                self.ventilation_on.astype(float) * self._attr('rated_input_power')
            )
        return self._energy_use


class VentilationSimulation(EnergyUseSimulation):
    ventilation_simulations: List[VentilationSystemSimulation]

    def __init__(
            self,
            spec: OpenBESSpecification,
            occupancy: OccupationSimulation = None,
            geometry: BuildingGeometry = None
    ):
        super().__init__(spec=spec)
        self.geometry = geometry or BuildingGeometry(spec=self.spec)
        self.occupancy = occupancy or OccupationSimulation(spec=self.spec, geometry=self.geometry)
        self.ventilation_simulations = []
        while True:
            system_number = len(self.ventilation_simulations) + 1
            if not hasattr(spec, f"ventilation_system{system_number}_rated_input_power"):
                break
            self.ventilation_simulations.append(
                VentilationSystemSimulation(
                    spec=spec,
                    system_number=system_number,
                    occupancy=self.occupancy,
                    geometry=self.geometry
                )
            )

    @property
    def air_supply_rate(self) -> 'Series[float]':
        """Total hourly air supply rate (m3/h/m2) from all ventilation systems.
        [Hourly simulation column JA]
        """
        if 'air_supply_rate' not in self._hours.columns:
            total_air_supply = Series([0.0] * len(self._hours), index=self._hours.index)
            for sim in self.ventilation_simulations:
                total_air_supply += sim.air_supply_rate
            self._hours['air_supply_rate'] = total_air_supply
        return self._hours['air_supply_rate']

    @property
    def energy_use(self) -> 'Series[float]':
        """Ventilation energy use in kWh for each hour of the year for each ENERGY_SOURCES.
        """
        return sum([x.energy_use for x in self.ventilation_simulations])


import math
from typing import Optional

from pandas import MultiIndex, DataFrame, Series

from ..types import (
    OpenBESSpecification,
    OCCUPATION_ZONES,
    get_zone_number,
    COMPASS_POINTS,
    FLOORS,
    ORIENTATIONS,
    THERMAL_BREAKS, HEAT_CAPACITY_CLASSES,
)

THERMAL_BREAK_TRANSMITTANCE = {
    THERMAL_BREAKS.Facade_ground: 0.54,
    THERMAL_BREAKS.Facade_intermediate: 0.60,
    THERMAL_BREAKS.Facade_roof: 0.44,
    THERMAL_BREAKS.Windows: 0.50,
    THERMAL_BREAKS.Shading: 0.80,
}

ZONAL_RECTANGLES = DataFrame(
    index=MultiIndex.from_product([list(FLOORS), list(OCCUPATION_ZONES)], names=['floor', 'zone'])
)

ORIENTATION_FACADE = DataFrame(
    index=MultiIndex.from_product([list(FLOORS), list(ORIENTATIONS)], names=['floor', 'orientation'])
)

COMPASS_POINT_FACADE = DataFrame(
    index=MultiIndex.from_product([list(FLOORS), list(COMPASS_POINTS)], names=['floor', 'compass_point'])
)

# Map whether each orientation is exposed to each compass point
EXPOSURES_MAP = DataFrame(
    index=MultiIndex.from_product([list(ORIENTATIONS), list(COMPASS_POINTS)], names=['orientation', 'compass_point'])
)

class Rectangle:
    def __init__(self, length: float, width: float):
        self.length = length
        self.width = width

    def __eq__(self, other):
        if not isinstance(other, Rectangle):
            return NotImplemented
        return self.length == other.length and self.width == other.width

    def compare(self, other):
        return f"Length: {self.length} vs {other.length}, Width: {self.width} vs {other.width}"

    @property
    def ratio(self):
        return self.length / self.width

    @property
    def area(self):
        return self.length * self.width


class BuildingGeometry:
    """
    Class to handle building geometry calculations based on OpenBESS specifications.
    Calculates equivalent rectangle, gross floor area, external vertical envelope area,
    window counts, window areas, and other geometry-related metrics.

    """
    spec: OpenBESSpecification
    _equivalent_rectangle: Rectangle
    _gross_floor_area: DataFrame
    _rectangles: DataFrame
    _orientation_facade: DataFrame
    _compass_point_facade: DataFrame
    _exposures: DataFrame
    _heat_transfer_rate_windows: float
    _heat_infiltration_opaque: float
    _heat_transfer_is: float
    _heat_transfer_ms: float
    _conditioned_floor_perimeters: Series
    _roof_projections: Series
    _roof_factor: float
    _conditioned_floor_area: float
    _building_mass_factor: float
    _heat_capacity_am: float
    _heat_capacity_cm: float

    def __init__(self, spec: OpenBESSpecification):
        self.spec = spec
        self._rectangles = ZONAL_RECTANGLES.copy()
        self._orientation_facade = ORIENTATION_FACADE.copy()
        self._compass_point_facade = COMPASS_POINT_FACADE.copy()
        self._exposures = EXPOSURES_MAP.copy()

    @property
    def heat_capacity_am(self) -> float:
        """Building heat capacity factor Am (m²).
        [Database cells AF5:AF10]
        Frequently appears as Am/Af e.g. in Hourly simulation AM91
        """
        if not hasattr(self, '_heat_capacity_am') or self._heat_capacity_am is None:
            if self.spec.heat_capacity in [
                HEAT_CAPACITY_CLASSES.Very_light,
                HEAT_CAPACITY_CLASSES.Light,
                HEAT_CAPACITY_CLASSES.Medium,
            ]:
                self._heat_capacity_am = 2.5
            elif self.spec.heat_capacity == HEAT_CAPACITY_CLASSES.Heavy:
                self._heat_capacity_am = 3.0
            elif self.spec.heat_capacity == HEAT_CAPACITY_CLASSES.Very_heavy:
                self._heat_capacity_am = 3.5
            elif self.spec.heat_capacity == HEAT_CAPACITY_CLASSES.Custom_value:
                self._heat_capacity_am = self.spec.parameters.advanced_heat_capacity_am
            else:
                raise ValueError(f"Unknown heat capacity class: {self.spec.heat_capacity}")
        return self._heat_capacity_am

    @property
    def heat_capacity_cm(self) -> float:
        """Building heat capacity Cm (kJ/K).
        [Database cells AG5:AG10]
        """
        if not hasattr(self, '_heat_capacity_cm') or self._heat_capacity_cm is None:
            if self.spec.heat_capacity == HEAT_CAPACITY_CLASSES.Very_light:
                self._heat_capacity_cm = 80_000.0
            elif self.spec.heat_capacity == HEAT_CAPACITY_CLASSES.Light:
                self._heat_capacity_cm = 110_000.0
            elif self.spec.heat_capacity == HEAT_CAPACITY_CLASSES.Medium:
                self._heat_capacity_cm = 165_000.0
            elif self.spec.heat_capacity == HEAT_CAPACITY_CLASSES.Heavy:
                self._heat_capacity_cm = 260_000.0
            elif self.spec.heat_capacity == HEAT_CAPACITY_CLASSES.Very_heavy:
                self._heat_capacity_cm = 370_000.0
            elif self.spec.heat_capacity == HEAT_CAPACITY_CLASSES.Custom_value:
                self._heat_capacity_cm = (
                        self.spec.parameters.heat_capacity_joule + self.spec.parameters.air_heat_capacity
                )
            else:
                raise ValueError(f"Unknown heat capacity class: {self.spec.heat_capacity}")
        return self._heat_capacity_cm

    @property
    def equivalent_rectangle(self) -> Rectangle:
        """Calculate the length and width of the equivalent rectangle of the building.
        [Inputs cells near B49]
    
        Returns:
            Rectangle: Equivalent rectangle for the input building geometry.
        """
        if not hasattr(self, '_equivalent_rectangle') or not isinstance(self._equivalent_rectangle, Rectangle):
            parameters = self.spec.parameters
            if (
                    parameters.courtyard_number == 0 and
                    parameters.open_courtyard_number_a1 == 0 and
                    parameters.open_courtyard_number_b1 == 0 and
                    parameters.open_courtyard_number_c1 == 0 and
                    parameters.open_courtyard_number_d1 == 0
            ):
                self._equivalent_rectangle = Rectangle(length=self.spec.building_length, width=self.spec.building_width)
            elif (
                    (parameters.courtyard_number != 0 and parameters.courtyard_length is None) or
                    (parameters.open_courtyard_number_a1 != 0 and parameters.open_courtyard_depth_a1 is None) or
                    (parameters.open_courtyard_number_b1 != 0 and parameters.open_courtyard_depth_b1 is None) or
                    (parameters.open_courtyard_number_c1 != 0 and parameters.open_courtyard_depth_c1 is None) or
                    (parameters.open_courtyard_number_d1 != 0 and parameters.open_courtyard_depth_d1 is None)
            ):
                raise ValueError("Courtyard dimensions must be provided if courtyard numbers are greater than zero.")
            else:
                length = (
                        self.spec.building_length +
                        float(parameters.courtyard_number or 0) * parameters.courtyard_length +
                        parameters.open_courtyard_depth_b1 * float(parameters.open_courtyard_number_b1 or 0) +
                        parameters.open_courtyard_depth_d1 * float(parameters.open_courtyard_number_d1 or 0)
                )
                width = (
                        self.spec.building_width +
                        float(parameters.courtyard_number or 0) * parameters.courtyard_width +
                        parameters.open_courtyard_depth_a1 * float(parameters.open_courtyard_number_a1 or 0) +
                        parameters.open_courtyard_depth_c1 * float(parameters.open_courtyard_number_c1 or 0)
                )
                self._equivalent_rectangle = Rectangle(length=length, width=width)
        return self._equivalent_rectangle

    def _get_gross_floor_area(self, row: Series) -> float:
        """[Reads values from Tool C58:G65]"""
        zone = row.name[1]
        z = get_zone_number(zone)
        if getattr(self.spec, f'condition_z{z}'):
            floor = row.name[0]
            return getattr(self.spec, f"{floor.value}_floor_area_z{z}") or 0.0
        return 0.0

    @property
    def gross_floor_areas(self) -> 'Series[float]':
        """Floor area of the building in square meters for each zone and floor.
        Af [Inputs cells C39:G45]
        """
        if 'gross_floor_area' not in self._rectangles.columns:
            self._rectangles['gross_floor_area'] = 0.0
            self._rectangles['gross_floor_area'] = self._rectangles.apply(self._get_gross_floor_area, axis=1)

        return self._rectangles['gross_floor_area']

    def get_gross_floor_area_for_floor(self, floor: FLOORS) -> float:
        """Calculate the total gross floor area for a specific floor in square meters.
        [Inputs cells I40:44]
        Args:
            floor (FLOORS): The floor to calculate the gross floor area for.

        Returns:
            float: Total gross floor area for the specified floor in square meters.
        """
        return self.gross_floor_areas.xs(floor, level='floor').sum()

    @property
    def gross_floor_area(self) -> float:
        """Calculate the total gross floor area of the building in square meters.
        A,G [Inputs cell I45]

        Returns:
            float: Total gross floor area in square meters.
        """
        return self.gross_floor_areas.sum()

    def _get_conditioned_floor_area(self, row: Series) -> float:
        z = get_zone_number(row.name[1])
        if not getattr(self.spec, f'condition_z{z}'):
            return 0.0
        return row['gross_floor_area'] * self.spec.parameters.nia_gba_ratio

    @property
    def conditioned_floor_areas(self) -> 'Series[float]':
        """Conditioned floor area of the building in square meters for each zone and floor.
        N45 [Inputs cell N45]
        """
        if 'conditioned_floor_area' not in self._rectangles.columns:
            self._rectangles['conditioned_floor_area'] = DataFrame(self.gross_floor_areas).apply(
                self._get_conditioned_floor_area, axis=1
            )
        return self._rectangles['conditioned_floor_area']

    def get_conditioned_floor_area_for_floor(self, floor: FLOORS) -> float:
        """Calculate the total conditioned floor area for a specific floor in square meters.
        [Inputs cells N40:44]
        Args:
            floor (FLOORS): The floor to calculate the conditioned floor area for.

        Returns:
            float: Total conditioned floor area for the specified floor in square meters.
        """
        return self.conditioned_floor_areas.xs(floor, level='floor').sum()

    @property
    def conditioned_floor_area(self) -> float:
        """Total conditioned floor area of the building in square meters.
        Af (in some places Af means gross, others conditioned)
        [Hourly simulation cell AM80]
        """
        if not hasattr(self, '_conditioned_floor_area') or self._conditioned_floor_area is None:
            self._conditioned_floor_area = self.conditioned_floor_areas.sum()
        return self._conditioned_floor_area

    def _get_external_vertical_envelope_area(self,
                                             spec: OpenBESSpecification,
                                             row: DataFrame,
                                             conditioned: bool = False
                                             ) -> float:
        """Calculate the total external vertical envelope area of the building in square meters.
    
        Args:
            row: DataFrame row with MultiIndex including 'floor' and 'orientation'.
            conditioned: If True, calculate conditioned area; if False, calculate gross area.
        Returns:
            float: Total external vertical envelope area in square meters.
        """
        # [Inputs cell C54]
        building_rectangular_ratio = self.equivalent_rectangle.ratio
        height = spec.floor_to_ceiling_height + spec.slab_thickness
        orientation = row.name[1]
        floor = row.name[0]
        if conditioned:
            floor_area = self.get_conditioned_floor_area_for_floor(floor)
        else:
            floor_area = self.get_gross_floor_area_for_floor(floor)

        if orientation in [ORIENTATIONS.Up, ORIENTATIONS.Down]:
            return math.sqrt(floor_area * building_rectangular_ratio) * height  # length
        else:
            return math.sqrt(floor_area / building_rectangular_ratio) * height  # width

    @property
    def external_vertical_envelope_gross_areas(self) -> 'Series[float]':
        """External vertical envelope gross area of the building in square meters for each orientation and floor.
        [Input cells C64:F69]
        """
        if 'external_vertical_envelope_gross_area' not in self._orientation_facade.columns:
            self._orientation_facade['external_vertical_envelope_gross_area'] = 0.0
            self._orientation_facade['external_vertical_envelope_gross_area'] = self._orientation_facade.apply(
                lambda row: self._get_external_vertical_envelope_area(
                    spec=self.spec,
                    row=row,
                    conditioned=False
                ),
                axis=1
            )
        return self._orientation_facade['external_vertical_envelope_gross_area']

    @property
    def external_vertical_envelope_conditioned_areas(self) -> 'Series[float]':
        """External vertical envelope conditioned area of the building in square meters for each orientation and floor.
        [Input cells C74:F79]
        """
        if 'external_vertical_envelope_conditioned_area' not in self._orientation_facade.columns:
            self._orientation_facade['external_vertical_envelope_conditioned_area'] = 0.0
            self._orientation_facade['external_vertical_envelope_conditioned_area'] = DataFrame(
                self._orientation_facade
            ).apply(
                lambda row: self._get_external_vertical_envelope_area(
                    spec=self.spec,
                    row=row,
                    conditioned=True
                ),
                axis=1
            )
        return self._orientation_facade['external_vertical_envelope_conditioned_area']

    def _get_window_count(self, row: Series):
        """
        Calculate the number of windows for a specific floor and orientation.

        Args:
            row: Series indexed by floor, orientation.
        """
        floor = row.name[0]
        orientation = row.name[1]
        floor_suffix = {
            FLOORS.Ground: "ground",
            FLOORS.First: "first",
            FLOORS.Second: "second",
            FLOORS.Third: "third",
            FLOORS.Fourth: "fourth",
        }[floor]
        orientation_suffix = {
            ORIENTATIONS.Up: "a1",
            ORIENTATIONS.Right: "b1",
            ORIENTATIONS.Down: "c1",
            ORIENTATIONS.Left: "d1",
        }[orientation]
        window_count_attr = f'window_number_{floor_suffix}_{orientation_suffix}'
        return getattr(self.spec, window_count_attr, 0) or 0

    @property
    def window_count(self) -> 'Series[int]':
        """Number of windows for each floor and orientation.

        [Inputs cells C90:F94] [Tool cells G72:J76]
        """
        if 'window_count' not in self._orientation_facade.columns:
            self._orientation_facade['window_count'] = 0
            self._orientation_facade['window_count'] = DataFrame(self._orientation_facade).apply(
                self._get_window_count,
                axis=1
            )
        return self._orientation_facade['window_count']

    @property
    def window_area_orientation(self) -> 'Series[float]':
        """Window area in square meters for each floor and orientation.
        """
        if 'window_area_orientation' not in self._orientation_facade.columns:
            if self.spec.window_height is None or self.spec.window_length is None:
                raise ValueError("Window height and length are required to model window area.")
            self._orientation_facade['window_area_orientation'] = (
                    self.window_count * self.spec.window_height * self.spec.window_length
            )
        return self._orientation_facade['window_area_orientation']

    @property
    def window_ratio(self) -> 'Series[float]':
        """The proportion of each vertical envelope taken up by windows for each floor and orientation.
        [Inputs cells H90:K94]
        """
        if 'window_ratio' not in self._orientation_facade.columns:
            # Assertions ensure we have the required columns calculated
            assert self.external_vertical_envelope_gross_areas is not None
            assert self.window_area_orientation is not None
            self._orientation_facade['window_ratio'] = self._orientation_facade.apply(
                lambda row: (
                    row['window_area_orientation'] / row['external_vertical_envelope_gross_area']
                    if row['external_vertical_envelope_gross_area'] != 0.0 else None
                ),
                axis=1
            )
            self._orientation_facade['window_ratio'] = self._orientation_facade['window_ratio'].fillna(0.0)
        return self._orientation_facade['window_ratio']

    @classmethod
    def get_facing_direction(cls, orientation_angle: float) -> COMPASS_POINTS:
        """Return the compass point that a given orientation angle faces towards."""
        orientation_angle = orientation_angle % 360
        if orientation_angle < 22.5:
            return COMPASS_POINTS.North
        if orientation_angle < (22.5 + 37.5):
            return COMPASS_POINTS.NorthEast
        if orientation_angle < (22.5 + 37.5 + 51):
            return COMPASS_POINTS.East
        if orientation_angle < (22.5 + 37.5 + 51 + 51):
            return COMPASS_POINTS.SouthEast
        if orientation_angle < (22.5 + 37.5 + 51 + 51 + 36):
            return COMPASS_POINTS.South
        if orientation_angle < (22.5 + 37.5 + 51 + 51 + 36 + 51):
            return COMPASS_POINTS.SouthWest
        if orientation_angle < (22.5 + 37.5 + 51 + 51 + 36 + 51 + 51):
            return COMPASS_POINTS.West
        if orientation_angle < (22.5 + 37.5 + 51 + 51 + 36 + 51 + 51 + 37.5):
            return COMPASS_POINTS.NorthWest
        return COMPASS_POINTS.North

    def get_compass_point_for_orientation(self, orientation: ORIENTATIONS) -> COMPASS_POINTS:
        """Return the compass point that a given orientation faces towards."""
        if orientation == ORIENTATIONS.Up:
            return self.get_facing_direction(self.spec.orientation_angle)
        if orientation == ORIENTATIONS.Right:
            return self.get_facing_direction(self.spec.orientation_angle + 90)
        if orientation == ORIENTATIONS.Down:
            return self.get_facing_direction(self.spec.orientation_angle + 180)
        return self.get_facing_direction(self.spec.orientation_angle + 270)

    def get_orientation_for_compass_point(self, compass_point: COMPASS_POINTS) -> Optional[ORIENTATIONS]:
        """Return the orientation that faces towards a given compass point."""
        for orientation in ORIENTATIONS:
            if self.get_compass_point_for_orientation(orientation) == compass_point:
                return orientation
        return None

    @property
    def conditioned_facade_areas(self):
        """The facade areas of the building in square meters by floor and compass point.
        [Inputs cells near M90]
        """
        if 'facade_areas' not in self._compass_point_facade.columns:
            assert self.external_vertical_envelope_conditioned_areas is not None
            self._compass_point_facade['facade_areas'] = 0.0
            for orientation in ORIENTATIONS:
                compass_point = self.get_compass_point_for_orientation(orientation)
                mask = self._compass_point_facade.index.get_level_values('compass_point') == compass_point
                self._compass_point_facade.loc[mask, 'facade_areas'] = (
                    self.external_vertical_envelope_conditioned_areas.xs(orientation, level='orientation')
                ).values
        return self._compass_point_facade['facade_areas']

    def _get_window_area(self, row: Series) -> float:
        """Calculate the window area for a specific floor and compass point.

        Args:
            row: Series indexed by floor, compass_point. Must have 'window_area_orientation' column.

        Returns:
            float: Window area in square meters.
        """
        floor = row.name[0]
        compass_point = row.name[1]
        orientation = self.get_orientation_for_compass_point(compass_point)
        if orientation is None:
            return 0.0
        assert self.window_ratio is not None
        assert self.external_vertical_envelope_conditioned_areas is not None
        r = self._orientation_facade.loc[(floor, orientation)]
        return r['window_ratio'] * r['external_vertical_envelope_conditioned_area']

    @property
    def window_areas(self) -> 'Series[float]':
        """The window area of the building in square meters by floor and compass point.
        [Inputs cells near M120]
        """
        if 'window_areas' not in self._compass_point_facade.columns:
            self._compass_point_facade['window_areas'] = 0.0
            self._compass_point_facade['window_areas'] = self._compass_point_facade.apply(
                self._get_window_area,
                axis=1
            )
        return self._compass_point_facade['window_areas']

    @property
    def window_area(self) -> float:
        """Total window area of the building in square meters.
        A,w
        """
        return self.window_areas.sum()


    @property
    def window_shading(self) -> 'Series[float]':
        """Shaded window factor for each floor.
    
        [Inputs cells AB120:124]
        """
        external_perimeter_rate = 0.75  # Constant, hardcoded in Inputs cell X115
        return (
                (self.window_areas.groupby(level='floor').sum() / self.spec.window_height) * external_perimeter_rate
        ).squeeze()

    @property
    def opaque_areas(self) -> 'Series[float]':
        """Opaque areas of the building in square meters by floor and compass point.
        """
        return self.conditioned_facade_areas - self.window_areas

    @property
    def opaque_area(self) -> float:
        """Total opaque area of the building in square meters.
        """
        return self.opaque_areas.sum()

    @property
    def building_mass_area(self) -> float:
        """Building mass area in m².
        Am  (sometimes Af????)
        """
        return self.heat_capacity_am * self.conditioned_floor_area

    @property
    def heat_transfer_rate_windows(self) -> float:
        """The heat transfer rate through windows (W/m²K).
        Htr_w [Hourly Simulation cell AR97]
        """
        if not hasattr(self, '_heat_transfer_rate_windows') or self._heat_transfer_rate_windows is None:
            correction_factor = self.spec.parameters.window_correction_factor
            window_area = self.window_area
            u_value = self.spec.uvalue_window
            if self.spec.thermal_bridge_shading:
                shading = self.window_shading * THERMAL_BREAK_TRANSMITTANCE[THERMAL_BREAKS.Shading]
            else:
                shading = 0.0
            self._heat_transfer_rate_windows = (
                    (window_area * u_value + shading) *
                    correction_factor /
                    self.conditioned_floor_area
            )
        return self._heat_transfer_rate_windows

    @property
    def conditioned_floor_perimeters(self) -> 'Series[float]':
        """Perimeter each floor in m.
        [Inputs cells W120:124]
        """
        if not hasattr(self, '_conditioned_floor_perimeters') or self._conditioned_floor_perimeters is None:
            data = []
            for i, floor in enumerate(FLOORS):
                floor_area = self.get_conditioned_floor_area_for_floor(floor)
                floor_length = math.sqrt(floor_area * self.equivalent_rectangle.ratio)
                floor_width = math.sqrt(floor_area / self.equivalent_rectangle.ratio)
                floor_exposure = 2 * floor_length + 2 * floor_width
                data.append(floor_exposure)
            self._conditioned_floor_perimeters = Series(data, index=list(FLOORS))

        return self._conditioned_floor_perimeters

    @property
    def roof_projections(self) -> 'Series[float]':
        """Roof area of the building in square meters by floor.
        [Inputs cells K120:124]
        """
        if not hasattr(self, '_roof_projections') or self._roof_projections is None:
            data = []
            for i, floor in enumerate(FLOORS):
                my_area = self.get_conditioned_floor_area_for_floor(floor)
                if floor == FLOORS.get_by_index(-1):
                    data.append(my_area)
                else:
                    area_above = self.get_conditioned_floor_area_for_floor(FLOORS.get_by_index(i + 1))
                    if my_area >= area_above:
                        data.append(my_area - area_above)
                    else:
                        data.append(0.0)
            self._roof_projections = Series(data, index=list(FLOORS))
        return self._roof_projections

    @property
    def roof_factor(self) -> float:
        """Roof factor for the building (roof area / building area).
        [Inputs cell M37]
        
        Used to convert between ground projection and actual roof area for slanted roofs.
        """
        if not hasattr(self, '_roof_factor') or self._roof_factor is None:
            roof_slope = self.spec.roof_angle  # [Inputs cell M32]
            horizontal = self.equivalent_rectangle.width / 2  # [Inputs cell M33]
            height = roof_slope * horizontal / 100  # [Inputs cell M34]
            slope = math.sqrt(horizontal**2 + height**2)  # [Inputs cell M35]
            roof_area = slope * 2 * self.equivalent_rectangle.length  # [Inputs cell M36]
            self._roof_factor = roof_area / self.equivalent_rectangle.area  # [Inputs cell M37]
        return self._roof_factor
    
    @property
    def conditioned_floor_projection(self) -> 'Series[float]':
        """Projection of each floor of the building in square meters by floor.
        For the ground floor, this is the conditioned floor area.
        For floors above the ground floor, this is the additional area compared to the floor below (i.e. the overhang).
        [Inputs cells L120:124]
        """
        data = []
        for i, floor in enumerate(FLOORS):
            my_area = self.get_conditioned_floor_area_for_floor(floor)
            if floor == FLOORS.Ground:
                data.append(my_area)
            else:
                area_below = self.get_conditioned_floor_area_for_floor(FLOORS.get_by_index(i - 1))
                if my_area <= area_below:
                    data.append(0.0)
                else:
                    data.append(my_area - area_below)
        return Series(data, index=list(FLOORS))

    @property
    def heat_infiltration_opaque(self) -> float:
        """Heat transfer rate through opaque envelope (W/m²K).
        Htr_opaque [Hourly Simulation cell AR81, Inputs C270]

        Calculated for the opaque elements of the building envelope including walls, roof, and ground floor by
        combining surface and linear transfer weighted by their respective areas and lengths.

        Surface transmission is area * U-value.

        Linear transmission is perimeter_length * ψ-value
        """
        if not hasattr(self, '_heat_infiltration_opaque') or self._heat_infiltration_opaque is None:
            floor_surface_area = self.conditioned_floor_projection.sum()  # [Inputs cell E259]
            wall_opaque_surface_area = self.conditioned_facade_areas.sum() - self.window_areas.sum()  # [C259]
            roof_surface_area = self.roof_projections.sum() * self.roof_factor  # [Inputs cell D259]

            floor_surface_transmission = floor_surface_area * self.spec.uvalue_floor  # [Inputs cell E260]
            wall_surface_transmission = wall_opaque_surface_area * self.spec.uvalue_facade  # [C260]
            roof_surface_transmission = roof_surface_area * self.spec.uvalue_roof  # [Inputs cell D260]

            # [Inputs cell C266]
            if self.spec.thermal_bridge_facade_ground:
                floor_transmittance = THERMAL_BREAK_TRANSMITTANCE[THERMAL_BREAKS.Facade_ground]
            else:
                floor_transmittance = 0.0
            if self.spec.thermal_bridge_facade_intermediate:
                other_floors_transmittance = THERMAL_BREAK_TRANSMITTANCE[THERMAL_BREAKS.Facade_intermediate]
            else:
                other_floors_transmittance = 0.0
            if self.spec.thermal_bridge_window:
                window_transmittance = THERMAL_BREAK_TRANSMITTANCE[THERMAL_BREAKS.Windows]
            else:
                window_transmittance = 0.0
            if self.spec.thermal_bridge_facade_roof:
                roof_transmittance = THERMAL_BREAK_TRANSMITTANCE[THERMAL_BREAKS.Facade_roof]
            else:
                roof_transmittance = 0.0

            floor_linear_transmission = self.conditioned_floor_perimeters[FLOORS.Ground] * floor_transmittance

            # intermediate floors have a more complex perimeter calculation for thermal bridging
            # [Inputs cells Y120:125]
            other_floors_thermal_bridge_perimeter = 0.0
            for i, floor in enumerate(FLOORS):
                if floor == FLOORS.Ground:
                    continue
                my_perimeter = self.conditioned_floor_perimeters[floor]
                prev_perimeter = self.conditioned_floor_perimeters[FLOORS.get_by_index(i - 1)]
                if prev_perimeter <= 0:
                    other_floors_thermal_bridge_perimeter += my_perimeter
                else:
                    other_floors_thermal_bridge_perimeter += my_perimeter * (my_perimeter / prev_perimeter)
            other_floors_linear_transmission = other_floors_thermal_bridge_perimeter * other_floors_transmittance
            # Windows get lumped in with other floors
            # [Inputs cells AA120:125]
            window_external_perimeter_rate = 0.75  # [Hardcoded in Inputs cell X115]
            window_perimeter = (
                    (self.window_areas.sum() / self.spec.window_height) +
                    (self.window_areas.sum() / self.spec.window_height * 2) * window_external_perimeter_rate
            )
            windows_linear_transmission = window_perimeter * window_transmittance

            # [Inputs cells D266, F266]
            wall_linear_transmission = (
                other_floors_linear_transmission +
                windows_linear_transmission
            )

            # [Inputs cells Z120:125]
            roof_perimeter = 0.0
            for i, floor in enumerate(FLOORS):
                if floor == FLOORS.get_by_index(-1):
                    roof_perimeter += self.conditioned_floor_perimeters[floor]
                else:
                    next_perimeter = self.conditioned_floor_perimeters[FLOORS.get_by_index(i + 1)]
                    my_perimeter = self.conditioned_floor_perimeters[floor]
                    roof_perimeter += (my_perimeter - next_perimeter)
            # [Inputs cell E266]
            roof_linear_transmission = roof_perimeter * roof_transmittance

            # [Calculated in Inputs C270]
            self._heat_infiltration_opaque = (
                    self.spec.parameters.floor_correction_factor *
                    (floor_surface_transmission + floor_linear_transmission) +
                    self.spec.parameters.facade_correction_factor *
                    (wall_surface_transmission + wall_linear_transmission) +
                    self.spec.parameters.roof_correction_factor * 
                    (roof_surface_transmission + roof_linear_transmission)
            )
        return self._heat_infiltration_opaque

    @property
    def heat_transfer_rate_opaque(self) -> float:
        """Heat transfer rate through opaque envelope (W/m²K).
        Htr_opaque [Hourly Simulation cell AR93]
        """
        return self.heat_infiltration_opaque / self.conditioned_floor_area

    @property
    def heat_transfer_ms(self) -> float:
        """Heat transfer coefficient between external air and external surface (W/m²K).
        Htr_ms [Hourly Simulation cell AR95] EN ISO 13790 12.2.2"""
        if not hasattr(self, '_heat_transfer_ms') or self._heat_transfer_ms is None:
            factor = 9.1  # Unnamed factor hardcoded in cell AR95 formula
            # Am or Am/Af is converted and reconverted several times, but ends up being just the lookup value
            Am = self.heat_capacity_am
            self._heat_transfer_ms = factor * Am
        return self._heat_transfer_ms

    @property
    def heat_transfer_is(self) -> float:
        """Heat transfer coefficient between internal air and internal surface (W/m²K).
        Htr_is [Hourly Simulation cell AR98] EN ISO 13790 12.2.2
        """
        if not hasattr(self, '_heat_transfer_is') or self._heat_transfer_is is None:
            Aat = 4.5  # [Hardcoded in Hourly Simulation cell AM84: EN ISO 13790, 7.2.2]
            Atot = Aat * self.conditioned_floor_area  # [AM85]
            Htr_is = 3.45 * Atot  # 3.45 hardcoded in [AR83]
            self._heat_transfer_is = Htr_is / self.conditioned_floor_area
        return self._heat_transfer_is



import logging
from typing import Dict

from pandas import DataFrame

from .base import EnergyUseSimulation
from .climate import ClimateSimulation
from .cooling import CoolingSimulation
from .geometry import BuildingGeometry
from .heating import HeatingSimulation
from .hot_water import HotWaterSimulation
from .lighting import LightingSimulation
from .occupancy import OccupationSimulation
from .ventilation import VentilationSimulation
from ..types import (
    OpenBESSpecification,
    ENERGY_USE_CATEGORIES,
    ENERGY_SOURCES,
)

logger = logging.getLogger(__name__)

class BuildingEnergySimulation(EnergyUseSimulation):
    """
    A building energy simulation takes a building specification and model parameters and produces a report
    on the energy use of the building.
    """

    def __init__(
            self,
            spec: OpenBESSpecification,
            hot_water: HotWaterSimulation = None,
            geometry: BuildingGeometry = None,
            occupancy: OccupationSimulation = None,
            lighting: LightingSimulation = None,
            ventilation: VentilationSimulation = None,
            climate: ClimateSimulation = None,
            cooling: CoolingSimulation = None,
            heating: HeatingSimulation = None,
    ):
        super().__init__(spec)
        self._standby_energy_use = self._energy_use.copy()
        self._standby_energy_use[ENERGY_SOURCES.Electricity] = (spec.building_standby_load * 12) / len(self._energy_use)
        self._other_energy_use = self._energy_use.copy()
        self._other_energy_use[ENERGY_SOURCES.Electricity] = (spec.other_electricity_usage * 12) / len(self._energy_use)
        self._other_energy_use[ENERGY_SOURCES.Natural_gas] = (spec.other_gas_usage * 12) / len(self._energy_use)
        self.hot_water = hot_water or HotWaterSimulation(self.spec)
        self.geometry = geometry or BuildingGeometry(self.spec)
        self.occupancy = occupancy or OccupationSimulation(self.spec, geometry=self.geometry)
        self.lighting = lighting or LightingSimulation(self.spec, occupancy=self.occupancy)
        self.ventilation = ventilation or VentilationSimulation(
            self.spec,
            occupancy=self.occupancy,
            geometry=self.geometry
        )
        self.climate = climate or ClimateSimulation(
            spec,
            geometry=self.geometry,
            occupancy=self.occupancy,
            lighting=self.lighting,
            ventilation=self.ventilation,
        )
        self.cooling = cooling or CoolingSimulation(
            spec,
            geometry=self.geometry,
            occupancy=self.occupancy,
            lighting=self.lighting,
            ventilation=self.ventilation,
            climate=self.climate,
        )
        self.heating = heating or HeatingSimulation(
            spec,
            geometry=self.geometry,
            occupancy=self.occupancy,
            lighting=self.lighting,
            ventilation=self.ventilation,
            climate=self.climate,
        )

    @property
    def energy_use_by_category(self) -> Dict[ENERGY_USE_CATEGORIES, DataFrame]:
        """Heating energy use in kWh for each hour of the year for each ENERGY_SOURCE for each ENERGY_USE_CATEGORY.
        """
        return {
                ENERGY_USE_CATEGORIES.Others: self._other_energy_use,
                ENERGY_USE_CATEGORIES.Building_standby: self._standby_energy_use,
                ENERGY_USE_CATEGORIES.Lighting: self.lighting.energy_use,
                ENERGY_USE_CATEGORIES.Hot_water: self.hot_water.energy_use,
                ENERGY_USE_CATEGORIES.Ventilation: self.ventilation.energy_use,
                ENERGY_USE_CATEGORIES.Cooling: self.cooling.energy_use,
                ENERGY_USE_CATEGORIES.Heating: self.heating.energy_use,
            }

    @property
    def energy_use(self) -> DataFrame:
        """Total energy use in kWh for each hour of the year for each ENERGY_SOURCE.
        """
        if self._energy_use.isna().any().any():
            self._energy_use.fillna(0, inplace=True)
            for category_use in self.energy_use_by_category.values():
                self._energy_use = self._energy_use.add(category_use, fill_value=0.0)
        return self._energy_use


from .types import (
    OpenBESSpecification,
    OpenBESParameters,
    ENERGY_SOURCES,
    COOLING_SYSTEM_TYPES,
    HEAT_CAPACITY_CLASSES,
    HEATING_SYSTEM_TYPES,
    LIGHTING_CONTROL,
    LIGHTING_BALLASTS,
    LIGHTING_TECHNOLOGIES,
    TERRAINS
)

HOLYWELL_HOUSE_SPEC = OpenBESSpecification(
    parameters=OpenBESParameters(
        temperature_tolerance=0
    ),
    appliances_load=1.0,
    biomass_annual=None,
    biomass_pellets_annual=None,
    building_area=None,
    building_height=5.5,
    building_length=43.41,
    building_name=None,
    building_standby_load=2321.2,
    building_type=None,
    building_width=14.13,
    condition_z1=True,
    condition_z2=True,
    condition_z3=True,
    condition_z4=True,
    condition_z5=False,
    cooling_system1_energy_efficifiency_ratio=2.8,
    cooling_system1_energy_source=ENERGY_SOURCES.Electricity,
    cooling_system1_nominal_capacity=96.0,
    cooling_system1_number=1,
    cooling_system1_off_time=17,
    cooling_system1_on_time=8,
    cooling_system1_sensible_nominal_capacity=75.0,
    cooling_system1_simultaneity_factor_canteen=1.0,
    cooling_system1_simultaneity_factor_common=0.0,
    cooling_system1_simultaneity_factor_office=1.0,
    cooling_system1_simultaneity_factor_other=0.0,
    cooling_system1_simultaneity_factor_teaching=1.0,
    cooling_system1_type=COOLING_SYSTEM_TYPES.Heat_pump,
    country=None,
    diesel_annual=None,
    electricity_annual=None,
    electricity_april=None,
    electricity_august=None,
    electricity_december=None,
    electricity_february=None,
    electricity_january=None,
    electricity_july=None,
    electricity_june=None,
    electricity_march=None,
    electricity_may=None,
    electricity_november=None,
    electricity_october=None,
    electricity_september=None,
    energy_generated=None,
    energy_used=None,
    first_floor_area_z1=494.85,
    first_floor_area_z2=None,
    first_floor_area_z3=None,
    first_floor_area_z4=55.65,
    first_floor_area_z5=None,
    floor_to_ceiling_height=2.90,
    fourth_floor_area_z1=None,
    fourth_floor_area_z2=None,
    fourth_floor_area_z3=None,
    fourth_floor_area_z4=None,
    fourth_floor_area_z5=None,
    gas_april=None,
    gas_august=None,
    gas_december=None,
    gas_february=None,
    gas_january=None,
    gas_july=None,
    gas_june=None,
    gas_march=None,
    gas_may=None,
    gas_november=None,
    gas_october=None,
    gas_september=None,
    ground_floor_area_z1=190.56,
    ground_floor_area_z2=129.28,
    ground_floor_area_z3=97.88,
    ground_floor_area_z4=132.76,
    ground_floor_area_z5=None,
    heat_capacity=HEAT_CAPACITY_CLASSES.Medium,
    heating_system1_efficiency_cop=0.86,
    heating_system1_energy_source=ENERGY_SOURCES.Natural_gas,
    heating_system1_nominal_capacity=48.0,
    heating_system1_number=2,
    heating_system1_off_time=None,
    heating_system1_on_time=None,
    heating_system1_simultaneity_factor_canteen=1.0,
    heating_system1_simultaneity_factor_common=0.2,
    heating_system1_simultaneity_factor_office=1.0,
    heating_system1_simultaneity_factor_other=0.0,
    heating_system1_simultaneity_factor_teaching=1.0,
    heating_system1_type=HEATING_SYSTEM_TYPES.Condensing_boiler,
    holiday=False,
    leakage_air_flow=None,
    leakage_air_flow_independent=0.45,
    lighting_control=LIGHTING_CONTROL.Manual,
    lighting_off_time=17,
    lighting_on_time=9,
    lighting_simultaneity_factor=0.2,
    lighting_system_ballast_z1=LIGHTING_BALLASTS.BE,
    lighting_system_ballast_z2=None,
    lighting_system_ballast_z3=None,
    lighting_system_ballast_z4=None,
    lighting_system_ballast_z5=None,
    lighting_system_ballast_z6=None,
    lighting_system_lamp_number_z1=4,
    lighting_system_lamp_number_z2=1,
    lighting_system_lamp_number_z3=None,
    lighting_system_lamp_number_z4=None,
    lighting_system_lamp_number_z5=None,
    lighting_system_lamp_number_z6=None,
    lighting_system_lamp_power_z1=18,
    lighting_system_lamp_power_z2=40,
    lighting_system_lamp_power_z3=None,
    lighting_system_lamp_power_z4=None,
    lighting_system_lamp_power_z5=None,
    lighting_system_lamp_power_z6=None,
    lighting_system_luminary_number_z1=35,
    lighting_system_luminary_number_z2=55,
    lighting_system_luminary_number_z3=None,
    lighting_system_luminary_number_z4=None,
    lighting_system_luminary_number_z5=None,
    lighting_system_luminary_number_z6=None,
    lighting_system_name_z1="First Floor",
    lighting_system_name_z2="Second Floor",
    lighting_system_name_z3=None,
    lighting_system_name_z4=None,
    lighting_system_name_z5=None,
    lighting_system_name_z6=None,
    lighting_system_operating_hours_z1=8,
    lighting_system_operating_hours_z2=8,
    lighting_system_operating_hours_z3=None,
    lighting_system_operating_hours_z4=None,
    lighting_system_operating_hours_z5=None,
    lighting_system_operating_hours_z6=None,
    lighting_system_similar_zone_number_z1=1,
    lighting_system_similar_zone_number_z2=1,
    lighting_system_similar_zone_number_z3=None,
    lighting_system_similar_zone_number_z4=None,
    lighting_system_similar_zone_number_z5=None,
    lighting_system_similar_zone_number_z6=None,
    lighting_system_simultaneity_factor_z1=0.7,
    lighting_system_simultaneity_factor_z2=0.7,
    lighting_system_simultaneity_factor_z3=None,
    lighting_system_simultaneity_factor_z4=None,
    lighting_system_simultaneity_factor_z5=None,
    lighting_system_simultaneity_factor_z6=None,
    lighting_system_tech_z1=LIGHTING_TECHNOLOGIES.FT_T8,
    lighting_system_tech_z2=LIGHTING_TECHNOLOGIES.LED,
    lighting_system_tech_z3=None,
    lighting_system_tech_z4=None,
    lighting_system_tech_z5=None,
    lighting_system_tech_z6=None,
    location=None,
    LPG_annual=None,
    max_building_occupation=150,
    meteorological_file="UK_Oxford_GBR_ENG_RAF.Benson.036580_TMYx.2007-2021.epw",
    natural_gas_annual=None,
    natural_ventilation_night=0.0,
    occupancy_close_canteen=14,
    occupancy_close_office=17,
    occupancy_close_teaching=17,
    occupancy_open_canteen=10,
    occupancy_open_office=8,
    occupancy_open_teaching=9,
    orientation_angle=165.0,
    other_electricity_usage=1136.0,
    other_gas_usage=0.0,
    roof_angle=35,
    schedule_april=True,
    schedule_august=True,
    schedule_december=True,
    schedule_february=True,
    schedule_friday=True,
    schedule_january=True,
    schedule_july=True,
    schedule_june=True,
    schedule_march=True,
    schedule_may=True,
    schedule_monday=True,
    schedule_november=True,
    schedule_october=True,
    schedule_saturday=False,
    schedule_september=True,
    schedule_sunday=False,
    schedule_thursday=True,
    schedule_tuesday=True,
    schedule_wednesday=True,
    second_floor_area_z1=None,
    second_floor_area_z2=None,
    second_floor_area_z3=None,
    second_floor_area_z4=52.93,
    second_floor_area_z5=None,
    setpoint_summer_day=21,
    setpoint_summer_night=29,
    setpoint_winter_day=22,
    setpoint_winter_night=18,
    slab_thickness=0.35,
    solar_external_shading_summer=0.9,
    solar_external_shading_winter=0.6,
    terrain_class=TERRAINS.Country,
    thermal_bridge_facade_ground=True,
    thermal_bridge_facade_intermediate=False,
    thermal_bridge_facade_roof=True,
    thermal_bridge_shading=False,
    thermal_bridge_window=False,
    third_floor_area_z1=None,
    third_floor_area_z2=None,
    third_floor_area_z3=None,
    third_floor_area_z4=None,
    third_floor_area_z5=None,
    typical_occupation=75,
    uvalue_facade=0.97,
    uvalue_floor=1.368,
    uvalue_roof=0.84,
    uvalue_window=2.5,
    ventilation_system1_airflow=300.0,
    ventilation_system1_energy_source=ENERGY_SOURCES.Electricity,
    ventilation_system1_heat_recovery_efficiency=0.75,
    ventilation_system1_off_time=14,
    ventilation_system1_on_time=10,
    ventilation_system1_rated_input_power=0.3,
    ventilation_system1_type=None,
    ventilation_system1_ventilated_area=100.0,
    water_demand=300.0,
    water_reference_temperature=60.0,
    water_supply_temperature=16.0,
    water_system_efficiency_cop=1.0,
    water_system_energy_source=ENERGY_SOURCES.Electricity,
    water_system_nominal_capacity=None,
    water_system_type=None,
    window_frame_factor=0.15,
    window_gvalue=0.65,
    window_height=1.6,
    window_length=2.4,
    window_number_first_a1=9,
    window_number_first_b1=3,
    window_number_first_c1=9,
    window_number_first_d1=None,
    window_number_fourth_a1=None,
    window_number_fourth_b1=None,
    window_number_fourth_c1=None,
    window_number_fourth_d1=None,
    window_number_ground_a1=9,
    window_number_ground_b1=3,
    window_number_ground_c1=9,
    window_number_ground_d1=None,
    window_number_second_a1=None,
    window_number_second_b1=None,
    window_number_second_c1=None,
    window_number_second_d1=None,
    window_number_third_a1=None,
    window_number_third_b1=None,
    window_number_third_c1=None,
    window_number_third_d1=None,
    zone_name_z1=None,
    zone_name_z2=None,
    zone_name_z3=None,
    zone_name_z4=None,
    zone_name_z5=None,
)



from importlib.resources import files
from math import atan
from pathlib import Path
from typing import Tuple
from numpy import nan, select, outer, array, maximum, isnan
from pvlib.iotools import read_epw
from pandas import DataFrame, Series
import os

from .base import HourlySimulation
from .geometry import BuildingGeometry
from .lighting import LightingSimulation
from .occupancy import OccupationSimulation
from .solar_irradiation import SolarIrradiationSimulation
from .ventilation import VentilationSimulation
from ..types import OpenBESSpecification, TERRAINS, ORIENTATIONS

# Optional dependencies
try:
    import numba
    USE_JIT = True
except (ImportError, ModuleNotFoundError):
    numba = None
    USE_JIT = False
try:
    import line_profiler
    USE_PROFILER = os.environ.get('LINE_PROFILE', '0') == '1'
except (ImportError, ModuleNotFoundError):
    line_profiler = None
    USE_PROFILER = False

def jit_if_available(_fn=None, **jit_kwargs):
    def decorator(fn):
        if not USE_JIT:
            return fn
        return numba.jit(**jit_kwargs)(fn)
    return decorator if _fn is None else decorator(_fn)

def profile_if_available(_fn=None, **profile_kwargs):
    """Profile if line_profiler is installed and enabled via LINE_PROFILE=1 environment variable."""
    def decorator(fn):
        if not USE_PROFILER:
            return fn
        return line_profiler.profile(**profile_kwargs)(fn)
    return decorator if _fn is None else decorator(_fn)

def profile_or_jit(_fn=None, jit_kwargs: dict = None, profile_kwargs: dict = None):
    """
    Can't have both JIT and line profiler at the same time, so this decorator applies one or the other if available.
    """
    def decorator(fn):
        if USE_PROFILER:
            return line_profiler.profile(**profile_kwargs)(fn)
        elif USE_JIT:
            return numba.jit(**jit_kwargs)(fn)
        else:
            return fn
    return decorator if _fn is None else decorator(_fn)
    

RELATIVE_HUMIDITY = 55.0  # Percentage

TERRAIN_VSITE_BY_VMETRO = {
    TERRAINS.Open: 1.0,
    TERRAINS.Country: 0.9,
    TERRAINS.Urban: 0.8
}

def get_available_epw_files() -> list[str]:
    """
    Returns a list of available EPW climate data files.
    """
    climate_data_dir = Path(str(files("openbes.simulations.climate_data")))
    return [
        f for f in os.listdir(climate_data_dir)
        if f.endswith('.epw')
    ]

@profile_or_jit(jit_kwargs={'nopython': True})
def _calculate_temperatures(
        prev_thermal_mass: float,
        Hc_nd: float,
        m: float,
        internal_heat_capacity_w: float,
        htr_1: float,
        htr_2: float,
        htr_3: float,
        Htr_em: float,
        dry_bulb_temp: float,
        Htr_is: float,
        Htr_ms: float,
        Htr_w: float,
        temp_st: float,
        internal_heat_adjusted: float,
        heat_transmission_by_ventilation: float,
        supply_air_temp: float,
):
    """Return a tuple of
    (m_tot, building_thermal_mass_t, internal_surface_temp, air_free_temp)
    for the given inputs.
    """
    m_tot = (
            m
            + Htr_em * dry_bulb_temp
            + htr_3 * (
                    temp_st +
                    Htr_w * dry_bulb_temp +
                    htr_1 * (
                            (internal_heat_adjusted + Hc_nd) / heat_transmission_by_ventilation +
                            supply_air_temp
                    )
            ) / htr_2
    )
    # Hourly variables are prev_thermal_mass, htr_3, m_tot
    building_thermal_mass_t = (
                                      prev_thermal_mass * (internal_heat_capacity_w - 0.5 * (htr_3 + Htr_em))
                                      + m_tot
                              ) / (internal_heat_capacity_w + 0.5 * (htr_3 + Htr_em))
    building_thermal_mass = (prev_thermal_mass + building_thermal_mass_t) / 2
    internal_surface_temp = (
                                    Htr_ms * building_thermal_mass
                                    + temp_st
                                    + Htr_w * dry_bulb_temp
                                    + htr_1
                                    * (
                                            supply_air_temp
                                            + (internal_heat_adjusted + Hc_nd)
                                            / heat_transmission_by_ventilation
                                    )
                            ) / (Htr_ms + Htr_w + htr_1)
    air_free_temp = (
                            Htr_is * internal_surface_temp
                            + heat_transmission_by_ventilation * supply_air_temp
                            + internal_heat_adjusted
                            + Hc_nd
                    ) / (Htr_is + heat_transmission_by_ventilation)
    return m_tot, building_thermal_mass_t, internal_surface_temp, air_free_temp

@profile_if_available
class ClimateSimulation(HourlySimulation):
    geometry: BuildingGeometry
    occupancy: OccupationSimulation
    lighting: LightingSimulation
    ventilation: VentilationSimulation
    _solar_irradiation: SolarIrradiationSimulation
    _epw_metadata: dict
    _epw_data: DataFrame
    _heating_and_cooling_degree_days: DataFrame
    _heat_infiltration_window: float
    _heat_infiltration_opaque: float
    _heat_transmission_by_infiltration: float
    _temp_change_demand: float
    _theta_st_partial: float

    def __init__(
            self,
            spec: OpenBESSpecification,
            geometry: BuildingGeometry = None,
            occupancy: OccupationSimulation = None,
            lighting: LightingSimulation = None,
            ventilation: VentilationSimulation = None,
    ):
        """
        Unlike other simulations, ClimateSimulation's Hourly data depend on the previous hourly data.
        This means that the class must calculate all hourly data in sequence, not just on demand.

        Consequently, the entire simulation is run in the __init__ method, which may take some time.
        """
        super().__init__(spec=spec)
        self.geometry = geometry or BuildingGeometry(spec=spec)
        self.occupancy = occupancy or OccupationSimulation(spec=spec, geometry=self.geometry)
        self.lighting = lighting or LightingSimulation(spec=spec, occupancy=self.occupancy)
        self.ventilation = ventilation or VentilationSimulation(
            spec=spec, geometry=self.geometry, occupancy=self.occupancy
        )
        # Pre-calculate all hour-dependent values in sequence
        n = len(self._hours)
        results = []
        row = None
        self._cache = self._populate_cache()
        for i in range(n):
            row = self._calculate_hour_row(i=i, prev_row=row)
            results.append(row)
        results_df = DataFrame(results, index=self._hours.index)
        self._hours = self._hours.join(results_df)
        # clear cache
        del self._cache

    def _populate_cache(self) -> dict:
        """
        Calculate the static values used in hourly row calculation that do not depend on the hour index.
        """
        # Make sure min/max set points are calculated
        assert self.set_point_temperature is not None
        # Window mask
        compass_points = sorted([self.geometry.get_compass_point_for_orientation(o) for o in ORIENTATIONS])
        window_mask = self.geometry.window_areas.index.get_level_values('compass_point').isin(compass_points)
        # [Hourly simulation cells KY105:LF105]
        window_area = self.geometry.window_areas[window_mask].groupby('compass_point').sum()
        # Solar heat windows reduction precaculated for all rows
        view_factor = self.spec.parameters.view_factor_to_sky_facade  # KY80:LF80
        hr = 5 * 0.9  # [Hardcoded in KY81:LF81 via Inputs G230]
        delta_theta_er = 11  # [Hardcoded in KY82: ISO 13790, 11.4.6]
        solar_heat_windows_reduction = DataFrame(
            outer(self.rse.values, window_area),
            index=self.rse.index,
            columns=window_area.index
        ) * view_factor * self.spec.uvalue_window * hr * delta_theta_er
        # Solar g_gl precalculated for all rows [Hourly simulation cell KX]
        solar_altitude = self.solar_irradiation.solar_altitude
        solar_g_gl = (
                             (-0.000003 * solar_altitude**3 + 0.0002 * solar_altitude**2 - 0.0053 * solar_altitude + 0.9986) *
                             self.spec.window_gvalue
                     ) * (solar_altitude > 0) * (solar_altitude < 90)
        # Asol for summer and winter [Hourly simulation cells KY107:LF108]
        solar_a_sol_w_winter = (
                (self.spec.solar_external_shading_winter * self.spec.parameters.shading_correction_factor) *
                (1 - self.spec.window_frame_factor) *
                window_area
        )
        solar_a_sol_w_summer = (
                (self.spec.solar_external_shading_summer * self.spec.parameters.shading_correction_factor) *
                (1 - self.spec.window_frame_factor) *
                window_area
        )
        # Solar area differs for summer and winter because of shading
        solar_area_summer = DataFrame(
            outer(solar_g_gl.values, solar_a_sol_w_summer.values),
            index=solar_heat_windows_reduction.index,
            columns=solar_a_sol_w_summer.index
        ) * 0.9  # [Hardcoded in KY:LF]
        solar_area_winter = DataFrame(
            outer(solar_g_gl.values, solar_a_sol_w_winter.values),
            index=solar_heat_windows_reduction.index,
            columns=solar_a_sol_w_winter.index
        ) * 1.0  # [Hardcoded in KY100]
        solar_radiation = self.solar_irradiation.solar_irradiation[compass_points]
        # Occupancy
        is_occupied = self.occupancy.occupancy['is_occupied']

        static = {
            "infiltration": (
                    self.spec.leakage_air_flow_independent * self.spec.parameters.infiltration_correction_factor
            ),
            "heat_capacity_air": (
                    self.spec.parameters.density_of_air
                    * self.spec.parameters.specific_heat_of_air
                    / 3.6
            ),
            "solar_heat_windows_reduction": array(solar_heat_windows_reduction),
            "solar_area_summer": array(solar_area_summer),
            "solar_area_winter": array(solar_area_winter),
            "solar_radiation": array(solar_radiation),
            # [Hourly simulation AR94]
            "Htr_em": 1 / (
                    (1 / self.geometry.heat_transfer_rate_opaque) -  # [AR93]
                    (1 / self.geometry.heat_transfer_ms)  # [AM94]
            ),
            "internal_heat_capacity_w": self.internal_heat_capacity / 3600,  # J/K to W/K [Hourly simulation AM93]
            "mass_factor_scaled": (
                    self.geometry.heat_capacity_am /
                    4.5  # [A_at hardcoded as 4.5 in Hourly Simulation cell AM84]
            ),
            # Arrayify some Series for performance
            "min_temp": array(self.set_point_temperature['min_temp_set_point'].values),
            "max_temp": array(self.set_point_temperature['max_temp_set_point'].values),
            "month": array(self.epw_data['month'].values),
            "day": array(self.epw_data['day'].values),
            "hour": array(self.epw_data['hour'].values),
            "solar_altitude": array(self.solar_irradiation.solar_altitude.values),
            "dry_bulb_temp": array(self.dry_bulb_temp.values),
            "wind_speed": array(self.wind_speed.values),
            "ventilation_air_supply_rate": array(self.ventilation.air_supply_rate.values),
            "solar_heat_opaque": array(self.solar_heat_opaque.values),
            "internal_heat": array(self.internal_heat.values),
            "internal_heat_adjusted": array(self.internal_heat_adjusted.values),
            "supply_air_temp": array(self.supply_air_temp.values),
            "is_occupied": array(is_occupied),
        }
        return static

    def _calculate_hour_row(self, i: int, prev_row: dict = None) -> dict:
        """
        Calculate all hour-dependent values for a given hour index.
        Optionally takes the previous row as input for recursive dependencies.
        """

        # Optionally use prev_row for values at i-1
        def get_prev(key, default=nan):
            return prev_row[key] if prev_row else default

        # --- Calculate all hour-dependent values ---
        min_temp = self._cache['min_temp'][i]
        max_temp = self._cache['max_temp'][i]

        """
        1. Night ventilation enabled [Hourly simulation column JL]

        Night ventilation is active between June 1st and September 1st inclusive, between sunset and dawn,
        if the air free temperature is above the dry bulb temperature.
        """
        prev_air_free_temp = get_prev("air_free_temp", 0.0)
        month = self._cache["month"][i]
        day = self._cache["day"][i]
        solar_altitude = self._cache["solar_altitude"][i]
        dry_bulb_temp = self._cache["dry_bulb_temp"][i]
        prev_dry_bulb_temp = 0.0 if i == 0 else self._cache["dry_bulb_temp"][i - 1]
        night_ventilation_enabled = (
            # [1st September 1am is coded to allow ventilation at Hourly Simulation cell JK5950]
                ((6 <= month < 9) or (month == 9 and day == 1 and self._cache['hour'][i] == 1))
                and (solar_altitude < 0)
                and (prev_air_free_temp > prev_dry_bulb_temp)
        )

        """
        2. Air flow base [Hourly simulation column JO]
        """
        threshold = 24
        on_hours = prev_air_free_temp >= threshold
        night_ventilation = (
                on_hours * self.spec.natural_ventilation_night * night_ventilation_enabled
        )
        air_flow_base = on_hours * night_ventilation + self._cache['infiltration']

        """
        3. Air flow dependent [Hourly simulation column JH]
        """
        qv_diff = 0.0  # [Hardcoded as blank in Hourly simulation column IU]
        q4pa = max(self.spec.parameters.leakage_air_flow_dependent, 0.0001)  # JE97
        Hstack = 10  # [Hardcoded in JE101: ISO 15242:2007. 6.7.1]
        air_set_temp_prev = get_prev("air_set_temp", 20.0)
        temp_diff = abs(dry_bulb_temp - air_set_temp_prev)
        qv_stack = max(0.0146 * q4pa * (((0.7 * Hstack) * temp_diff)**0.667), 0.001)
        dcp = 0.75  # [Hardcoded in JE103]
        vsite_by_vmetro = TERRAIN_VSITE_BY_VMETRO[self.spec.terrain_class]  # JE104
        qv_wind = (
                0.0769
                * q4pa
                * (dcp * (vsite_by_vmetro * self._cache['wind_speed'][i])**2)**0.667
        )
        qv_sw = max(qv_stack, qv_wind) + (0.14 * qv_stack * qv_wind / q4pa)
        qv_infred = max(
            qv_sw,
            (
                    qv_stack * abs(qv_diff / 2)
                    + qv_wind * 2 * abs(qv_diff / 3) / (qv_stack + qv_wind)
            ),
        )
        air_flow_dependent = qv_diff + qv_infred

        """
        4. Air flow [Hourly simulation column JZ]
        """
        air_flow = (
                self._cache["ventilation_air_supply_rate"][i]
                + air_flow_dependent
                + air_flow_base
        )

        """
        5. Heat transmission by ventilation [Hourly Simulation column AL]
        """
        heat_transmission_by_ventilation = self._cache['heat_capacity_air'] * air_flow

        """
        6. Htr_1 [Hourly Simulation column AM]
        """
        htr_1 = 1 / (
                1 / heat_transmission_by_ventilation + 1 / self.geometry.heat_transfer_is
        )

        """
        7. Htr_2 [Hourly Simulation column AN]
        """
        htr_2 = htr_1 + self.geometry.heat_transfer_rate_windows

        """
        8. Htr_3 [Hourly Simulation column AO]
        """
        htr_3 = 1 / ((1 / htr_2) + (1 / self.geometry.heat_transfer_ms))

        """
        9. Solar heat windows [Hourly Simulation columns KY:LF]
        """
        kx116 = 22  # Hardcoded in Hourly Simulation cell KX116
        Fsh_ob_overhand = 1.0  # LF87 [Hardcoded via 0 in Inputs N108:U108]
        Fsh_ob_fin = 1.0  # LF90 [Hardcoded via 0 in Inputs N110:U110]
        Fsh_ob_horizon = 1.0  # LF93 [Hardcoded via 0 in Inputs N112:U112]
        solar_lost_through_windows = 0.0  # [Hardcoded in KY84]
        # LF95: ISO 13790, 11.4.4
        Fsh_ob_horizon = Fsh_ob_overhand * Fsh_ob_fin * Fsh_ob_horizon - solar_lost_through_windows
        solar_radiation = self._cache["solar_radiation"][i]
        if prev_air_free_temp < kx116:
            solar_area = self._cache["solar_area_winter"][i]
        else:
            solar_area = self._cache["solar_area_summer"][i]

        solar_heat_base = solar_area * solar_radiation
        solar_heat_windows = (
                (Fsh_ob_horizon * solar_heat_base) -
                self._cache['solar_heat_windows_reduction'][i]
        )
        solar_heat_windows = solar_heat_windows * (solar_heat_windows > 0.0)
        solar_heat_windows = solar_heat_windows.sum()

        """
        10. Solar heat [Hourly Simulation column AJ, KO]
        """
        solar_heat_opaque = self._cache['solar_heat_opaque'][i]
        conditioned_floor_area = self.geometry.conditioned_floor_area
        solar_heat = (solar_heat_windows + solar_heat_opaque) / conditioned_floor_area

        """
        11. m [Hourly Simulation column AR]
        """
        internal_heat = self._cache['internal_heat'][i]
        internal_heat_adjusted = self._cache['internal_heat_adjusted'][i]
        m = self._cache['mass_factor_scaled'] * (
                0.5 * internal_heat +
                (
                        (solar_heat_windows + solar_heat_opaque) /
                        conditioned_floor_area
                )
        )

        """
        12. temp_st [Hourly Simulation column AS]
        """
        temp_st = self.theta_st_partial * (
                0.5 * internal_heat + solar_heat
        )

        """
        13. m_tot [Hourly Simulation column AT]
        14. Building thermal mass, Building thermal mass@t [Hourly Simulation columns AW, AV]
        15. Internal surface temp [Hourly Simulation column AX]
        16. Air free temp [Hourly Simulation column AY]
        """
        # m_tot parameters
        supply_air_temp = self._cache['supply_air_temp'][i]
        Htr_is = self.geometry.heat_transfer_is
        Hc_nd = 0.0  # Heating/cooling need [Hardcoded in AR111]
        Htr_em = self._cache["Htr_em"]
        # building_thermal_mass parameters
        starting_thermal_mass = 17.4  # [Hardcoded in Hourly Simulation cell AV117]
        internal_heat_capacity_w = self._cache["internal_heat_capacity_w"]
        prev_thermal_mass = get_prev("building_thermal_mass_t", starting_thermal_mass)

        # semi-curry _calculate_temperatures function now all fixed parameters are known
        def calculate_temperatures(hc_need: float, prev_tm: float) -> Tuple[float, float, float, float]:
            """Returns (m_tot, building_thermal_mass_t, internal_surface_temp, air_free_temp)"""
            return _calculate_temperatures(
                Hc_nd=hc_need,
                prev_thermal_mass=prev_tm,
                m=m,
                dry_bulb_temp=dry_bulb_temp,
                htr_1=htr_1,
                htr_2=htr_2,
                htr_3=htr_3,
                temp_st=temp_st,
                internal_heat_adjusted=internal_heat_adjusted,
                heat_transmission_by_ventilation=heat_transmission_by_ventilation,
                supply_air_temp=supply_air_temp,
                Htr_em=Htr_em,
                Htr_is=Htr_is,
                Htr_ms=9.1 * self.geometry.heat_capacity_am,  # [Hardcoded in Hourly Simulation cell AR95]
                Htr_w=self.geometry.heat_transfer_rate_windows,
                internal_heat_capacity_w=internal_heat_capacity_w,
            )

        m_tot, building_thermal_mass_t, internal_surface_temp, air_free_temp = calculate_temperatures(
            hc_need=Hc_nd,
            prev_tm=prev_thermal_mass
        )
        building_thermal_mass = (prev_thermal_mass + building_thermal_mass_t) / 2

        """
        17. Air free temp with heating/cooling to meet set point, heating/cooling need = 0 [Hourly Simulation column BK]
        """
        # Air free temps HC_0 and HC_10 are single-hour forward simulations if heating is on/off.
        #  They use the simulated temperature for the previous hour from the full HC_actual model.
        prev_thermal_mass_hc_actual = get_prev('building_thermal_mass_hc_actual_t', starting_thermal_mass)
        _, _, _, air_free_temp_hc_0 = calculate_temperatures(hc_need=0.0, prev_tm=prev_thermal_mass_hc_actual)

        """
        18. Air set temp [Hourly Simulation column CE]
        """
        if air_free_temp_hc_0 < max_temp:
            air_set_temp = max_temp
        elif air_free_temp_hc_0 > min_temp:
            air_set_temp = min_temp
        else:
            air_set_temp = air_free_temp_hc_0

        """
        19. Air temp with heating/cooling need = 10W/m2 [Hourly Simulation column BY]
        """
        # [HC,nd hardcoded as 10 in cell BP114]
        _, _, _, air_free_temp_hc_10 = calculate_temperatures(hc_need=10.0, prev_tm=prev_thermal_mass_hc_actual)

        """
        20. Building thermal mass/@t with heating/cooling need = actual need [Hourly Simulation columns CR, CQ]
        21. Air temp with heating/cooling need = actual need [Hourly Simulation column CT]
        Note: The heating/cooling need is an interpretation of the Excel spreadsheet's winter/summer set point temps.
        22. Heating/cooling need [Hourly Simulation column DB]
        """
        # [Hourly simulation column CK]
        is_occupied = self._cache['is_occupied'][i]
        needs_heating = air_free_temp_hc_0 < max_temp
        needs_cooling = air_free_temp_hc_0 > min_temp
        if is_occupied and (needs_heating or needs_cooling):
            heating_cooling_demand = (
                    10 * (air_set_temp - air_free_temp_hc_0) / (air_free_temp_hc_10 - air_free_temp_hc_0)
            )
        else:
            heating_cooling_demand = 0.0
        _, building_thermal_mass_hc_actual_t, _, air_free_temp_hc_actual = calculate_temperatures(
            hc_need=heating_cooling_demand,
            prev_tm=prev_thermal_mass_hc_actual
        )

        building_thermal_mass_hc_actual = (prev_thermal_mass_hc_actual + building_thermal_mass_hc_actual_t) / 2

        # --- Return all calculated values as a dict ---
        values = {
            "night_ventilation_enabled": night_ventilation_enabled,
            "air_flow_dependent": air_flow_dependent,
            "air_flow": air_flow,
            "heat_transmission_by_ventilation": heat_transmission_by_ventilation,
            "htr_1": htr_1,
            "htr_2": htr_2,
            "htr_3": htr_3,
            "solar_heat_windows": solar_heat_windows,
            "solar_heat": solar_heat,
            "m": m,
            "temp_st": temp_st,
            "m_tot": m_tot,
            "building_thermal_mass": building_thermal_mass,
            "building_thermal_mass_t": building_thermal_mass_t,
            "internal_surface_temp": internal_surface_temp,
            "air_free_temp": air_free_temp,
            "air_set_temp": air_set_temp,
            "air_free_temp_hc_0": air_free_temp_hc_0,
            "air_free_temp_hc_10": air_free_temp_hc_10,
            "building_thermal_mass_hc_actual": building_thermal_mass_hc_actual,
            "building_thermal_mass_hc_actual_t": building_thermal_mass_hc_actual_t,
            "air_free_temp_hc_actual": air_free_temp_hc_actual,
            "heating_cooling_demand": heating_cooling_demand,
        }

        if i == 1 and any(isnan(v) for v in values.values()):
            raise ValueError(f"NaN values:\n{DataFrame([values])}")

        return values

    @property
    def set_point_temperature(self) -> DataFrame:
        """Set point temperatures for each hour of the year.

        DataFrame with columns ['min_temp_set_point', 'max_temp_set_point'] for each hour.

        The set point temperatures provide a minimum and maximum temperature for comfortable building occupation.
        This is given by specified target temperatures with an optional tolerance.
        """
        if 'min_temp_set_point' not in self._hours.columns or 'max_temp_set_point' not in self._hours.columns:
            tolerance = max(self.spec.parameters.temperature_tolerance or 0.0, 0.0)
            self._hours['max_temp_set_point'] = self._hours['is_daytime'].apply(
                lambda x: (self.spec.setpoint_winter_day - tolerance) if x else (self.spec.setpoint_winter_night - tolerance)
            )
            self._hours['min_temp_set_point'] = self._hours['is_daytime'].apply(
                lambda x: (self.spec.setpoint_summer_day + tolerance) if x else (self.spec.setpoint_summer_night + tolerance)
            )
        return self._hours[['min_temp_set_point', 'max_temp_set_point']]

    @property
    def epw_data(self) -> DataFrame:
        """DataFrame with EPW climate data for the specified location.
        """
        if not hasattr(self, '_epw_data') or self._epw_data is None:
            file_name = self.spec.meteorological_file
            path = Path(str(files("openbes.simulations.climate_data") / file_name))
            self._epw_data, self._epw_metadata = read_epw(path)
            self._hours['wind_speed'] = array(self._epw_data['wind_speed'])
        return self._epw_data

    @property
    def epw_metadata(self) -> dict:
        """Dict with EPW metadata for the specified location."""
        if not hasattr(self, '_epw_metadata') or self._epw_metadata is None:
            assert self.epw_data is not None  # Trigger loading of EPW data and metadata
        return self._epw_metadata

    @property
    def solar_irradiation(self) -> SolarIrradiationSimulation:
        """Solar irradiation simulation for the building."""
        if not hasattr(self, '_solar_irradiation') or self._solar_irradiation is None:
            self._solar_irradiation = SolarIrradiationSimulation(
                epw_data=self.epw_data,
                epw_metadata=self.epw_metadata,
            )
        return self._solar_irradiation

    def get_heating_and_cooling_degrees_days(self, base_temperature: float = 18.0) -> DataFrame:
        """Calculate heating degree days for each month based on dry bulb temperature.
        Args:
            base_temperature (float): The base temperature for heating degree days calculation. Generally 65°F (18°C).
        Returns:
            DataFrame: Heating degree days for each month.
        """
        epw = self.epw_data
        days = epw[['month', 'day', 'temp_air']].copy()
        days = days.groupby(['month', 'day']).agg(lambda x: (x.max() + x.min()) / 2).reset_index(['month', 'day'])
        days['day'] = days.index + 1
        days = days.set_index(['day'])
        days['heating_degree_day'] = days['temp_air'].apply(lambda x: max(0, base_temperature - x))
        days['cooling_degree_day'] = days['temp_air'].apply(lambda x: max(0, x - base_temperature))
        days = days.drop(columns=['temp_air'])
        return days

    @property
    def htr_1(self) -> 'Series[float]':
        """Hourly heat transfer rate 1?????? in kW/K.
        Htr_1 [Hourly Simulation column AM]
        """
        return self._hours['htr_1']

    @property
    def htr_2(self) -> 'Series[float]':
        """Hourly heat transfer rate 2?????? in kW/K.
        Htr_2 [Hourly Simulation column AN]
        """
        return self._hours['htr_2']

    @property
    def htr_3(self) -> 'Series[float]':
        """Hourly heat transfer rate 3?????? in kW/K.
        Htr_3 [Hourly Simulation column AO]
        """
        return self._hours['htr_3']

    @property
    def internal_surface_temp(self) -> 'Series[float]':
        """Hourly internal surface temperature in degrees C.
        Ѳs [Hourly Simulation column AX]
        """
        return self._hours['internal_surface_temp']

    @property
    def theta_st_partial(self) -> float:
        """?????#
        Ѳst [Hourly simulation cell AR103]

        Eq. C.3 (partial)
        """
        if not hasattr(self, '_theta_st_partial') or self._theta_st_partial is None:
            A_at = 4.5  # [Hardcoded in Hourly Simulation cell AM84]
            self._theta_st_partial = (
                    1 -
                    (self.geometry.heat_capacity_am / A_at) -
                    (self.geometry.heat_transfer_rate_windows / (9.1 * A_at))
            )
        return self._theta_st_partial

    @property
    def temp_st(self) -> 'Series[float]':
        """Hourly temperature for ????? in degrees C.
        Ѳst [Hourly Simulation column AS]
        """
        return self._hours['temp_st']

    @property
    def internal_heat_capacity(self) -> float:
        """Calculate the internal heat capacity of the building in J/K.
        [Inputs cell C275]
        """
        return (
                self.spec.parameters.heat_capacity_correction_factor *
                self.geometry.heat_capacity_cm
        )

    @property
    def m(self) -> 'Series[float]':
        """Hourly m?????? in W/m2.
         Φm [Hourly Simulation column AR]
        """
        return self._hours['m']

    @property
    def m_tot(self) -> 'Series[float]':
        """Hourly m_tot?????? in W/m2.
        [Hourly Simulation column AT]
        """
        return self._hours['m_tot']

    @property
    def building_thermal_mass(self) -> 'Series[float]':
        """Hourly building thermal mass in degrees C.
        Ѳm [Hourly Simulation column AW]
        """
        return self._hours['building_thermal_mass']

    @property
    def building_thermal_mass_hc_actual(self) -> 'Series[float]':
        """Building thermal mass with heating/cooling need = actual need
        Ѳm [Hourly Simulation column CQ]
        """
        return self._hours['building_thermal_mass_hc_actual']

    @property
    def supply_air_temp(self) -> 'Series[float]':
        """Hourly supply air temperature in degrees C.
        Ѳsup = Te [Hourly simulation column AG]
        """
        if 'supply_air_temp' not in self._hours.columns:
            self._hours['supply_air_temp'] = array(self.epw_data['temp_air'])
        return self._hours['supply_air_temp']

    @property
    def relative_humidity(self) -> 'Series[float]':
        """Relative humidity for each hour of the year.
        """
        if 'relative_humidity' not in self._hours.columns:
            relative_humidity = self.epw_data['relative_humidity']
            relative_humidity.index = self._hours.index
            self._hours['relative_humidity'] = relative_humidity
        return self._hours['relative_humidity']

    @property
    def wet_bulb_temp(self) -> 'Series[float]':
        """Wet bulb temperature for each hour of the year.
        Tw [Hourly Simulation column K]
        """
        if 'wet_bulb_temp' not in self._hours.columns:
            df = self.relative_humidity.to_frame()
            df['temp_air'] = list(self.epw_data['temp_air'])
            self._hours['wet_bulb_temp'] = df.apply(
                lambda row: row['temp_air'] * \
                            atan(0.151977 * (row['relative_humidity'] + 8.313659)**0.5) + \
                            atan(row['temp_air'] + row['relative_humidity']) - \
                            atan(row['relative_humidity'] - 1.676331) + \
                            0.00391838 * (row['relative_humidity']**1.5) * atan(0.023101 * row['relative_humidity']) \
                            - 4.686035,
                axis=1
            )
        return self._hours['wet_bulb_temp']

    @property
    def dry_bulb_temp(self) -> 'Series[float]':
        """Dry bulb temperature for each hour of the year.
        [Hourly simulation column I]
        """
        if 'dry_bulb_temp' not in self._hours.columns:
            self._hours['dry_bulb_temp'] = array(self.epw_data['temp_air'])
        return self._hours['dry_bulb_temp']

    @property
    def night_ventilation_enabled(self) -> 'Series[bool]':
        """Whether night ventilation is active for each hour of the year.
        [Hourly simulation column JL]

        Night ventilation is active between June 1st and August 31st inclusive, between sunset and dawn,
        if the air free temperature is above the dry bulb temperature.

        Note: The spreadsheet has night ventilation enabled on September 1st at 1am, which may be a mistake.
         We reproduce this behaviour here for consistency.
        """
        return self._hours['night_ventilation_enabled']

    @property
    def air_flow_base(self) -> 'Series[float]':
        """Hourly base air flow in m3/h/m2.
        qv,base [Hourly simulation column JO]

        Base airflow is the infiltration airflow independent of other variables.

        Calculated by Q4Pa + night ventilation (qv,inf + qv,NV)
        """
        if 'air_flow_base' not in self._hours.columns:
            infiltration = self.spec.leakage_air_flow_independent * self.spec.parameters.infiltration_correction_factor
            threshold = 24  # [Hardcoded in Hourly simulation cell JK116]
            on_hours = self.air_free_temp >= threshold  # JM
            # [Hourly simulation column JM]
            night_ventilation = (
                    on_hours *
                    self.spec.natural_ventilation_night *
                    self.night_ventilation_enabled
            )
            self._hours['air_flow_base'] = on_hours * night_ventilation + infiltration
        return self._hours['air_flow_base']

    @property
    def air_set_temp(self) -> 'Series[float]':
        """Hourly air set temperature in degrees C.
        Ѳair,set [Hourly simulation column CE]

        This is the temperature that the heating/cooling systems will try to attain.
        """
        return self._hours['air_set_temp']

    @property
    def wind_speed(self) -> 'Series[float]':
        """Hourly wind speed in m/s.
        [Hourly simulation column W]
        """
        if 'wind_speed' not in self._hours.columns:
            self._hours['wind_speed'] = array(self.epw_data['wind_speed'])
        return self._hours['wind_speed']

    @property
    def air_flow_dependent(self) -> 'Series[float]':
        """Hourly air flow in m3/h/m2, dependent on other variables.
        qv,inf [Hourly simulation column JH]
        """
        return self._hours['air_flow_dependent']

    @property
    def air_flow(self) -> 'Series[float]':
        """Hourly air flow in m3/h/m2.
        qv,tot [Hourly simulation column JZ]

        Total airflow is
        air infiltration adjusted for other variables +
        air infiltration base +
        mechanical supply 1 +
        mechanical supply 2
        """
        return self._hours['air_flow']

    @property
    def heat_transmission_by_ventilation(self) -> 'Series[float]':
        """Calculate the heat transmission by ventilation in kW/K.
        Hve [Hourly Simulation column AL]

        Heat transfer of ventilation (Hve, W/m2 K) is calculated according to
        Eq. (5). It is based on total air flow due to leakage and ventilation
        airflow (qve), and supply air temperature (Ѳsup).
        """
        return self._hours['heat_transmission_by_ventilation']

    @property
    def heat_infiltration_window(self) -> float:
        """Calculate the heat transmission by infiltration through windows in kW/K.
        Htr,w
        """
        if not hasattr(self, '_heat_infiltration_window') or self._heat_infiltration_window is None:
            raise NotImplementedError
        return self._heat_infiltration_window

    @property
    def heat_transmission_by_infiltration(self) -> float:
        """Calculate the heat transmission by infiltration in kW/K.
        Htr
        """
        if not hasattr(self, '_heat_transmission_by_infiltration') or self._heat_transmission_by_infiltration is None:
            raise NotImplementedError
        return self._heat_transmission_by_infiltration

    @property
    def internal_heat_from_occupants(self) -> 'Series[float]':
        """Hourly internal heat gains from occupants in W/m2.
        ϕint,oc [Hourly Simulation column KI]
        """
        if 'internal_heat_from_occupants' not in self._hours.columns:
            self._hours['internal_heat_from_occupants'] = (
                    self.occupancy.occupancy['occupancy_ratio'] *
                    self.occupancy.metabolic_rate_per_m2
            )
        return self._hours['internal_heat_from_occupants']

    @property
    def internal_heat_from_appliances(self) -> 'Series[float]':
        """Hourly internal heat gains from appliances in W/m2.
        ϕint,ap [Hourly Simulation column KJ]
        """
        if 'internal_heat_from_appliances' not in self._hours.columns:
            # Inputs cell C144, Table G.11 ISO 13790
            appliance_W_per_m2 = self.spec.appliances_load * self.spec.parameters.appliance_on_off
            self._hours['internal_heat_from_appliances'] = (
                    self.occupancy.occupancy['occupancy_ratio'] *
                    appliance_W_per_m2
            )
        return self._hours['internal_heat_from_appliances']

    @property
    def internal_heat_from_lighting(self) -> 'Series[float]':
        """Hourly internal heat gains from lighting in W/m2.
        ϕint,l [Hourly Simulation column KK, KQ]

        Lighting heat generation is modelled using a constant standby (parasitic) output (Wpc) and
        an occupancy-scaled output (Wli).
        """
        if 'internal_heat_from_lighting' not in self._hours.columns:
            ratio = array(self.lighting.lighting_ratio)
            heat = array(self.lighting.lighting_heat)
            parasitic = array(self.lighting.parasitic_heat)
            self._hours['internal_heat_from_lighting'] = (
                    (ratio * heat) + parasitic
            )
        return self._hours['internal_heat_from_lighting']

    @property
    def internal_heat(self) -> 'Series[float]':
        """Hourly internal heat gains in W/m2.
        ϕint [Hourly Simulation column AI; KL]

        Internal heat gains are the sum of internal heat gains from occupants, appliances, and lighting.
        """
        if 'internal_heat' not in self._hours.columns:
            # calculate prerequisites
            self._hours['internal_heat'] = (
                    self.internal_heat_from_occupants +
                    self.internal_heat_from_appliances +
                    self.internal_heat_from_lighting
            )
        return self._hours['internal_heat']

    @property
    def internal_heat_adjusted(self) -> 'Series[float]':
        """Hourly adjusted internal heat gains in W/m2.""
        ϕia [Hourly Simulation column AQ]

        Adjusted internal heat gains are the internal heat gains multiplied by an adjustment factor.
        """
        if "internal_heat_adjusted" not in self._hours.columns:
            adjustment_factor = 0.5  # Hardcoded in spreadsheet column AQ
            self._hours['internal_heat_adjusted'] = self.internal_heat * adjustment_factor
        return self._hours['internal_heat_adjusted']

    @property
    def air_free_temp(self) -> 'Series[float]':
        """Hourly air free temperature with no heating or cooling adjustments.
        Ѳair,0 [Hourly Simulation column AY]

        Calculated by considering:
        - Internal surface temperature and its heat transfer rate to air
        - Heat transmission by ventilation and supply air temperature
        - Internal heat gains (adjusted)
        - HC_nd (assumed to be 0)
        and dividing by the total heat transfer rates to air (from surfaces and ventilation).

        This produces a weighted sum of these temperature influences to estimate the air free temperature.
        """
        return self._hours['air_free_temp']

    @property
    def air_free_temp_hc_0(self) -> 'Series[float]':
        """Hourly air free temperature with heating/cooling adjustments set to 0W/m2.
        Ѳair,0 [Hourly Simulation column BK]
        """
        return self._hours['air_free_temp_hc_0']

    @property
    def air_free_temp_hc_10(self) -> 'Series[float]':
        """Hourly air free temperature with heating/cooling adjustments set to 10W/m2.
        Ѳair,10 [Hourly Simulation column BY]
        """
        return self._hours['air_free_temp_hc_10']

    @property
    def air_free_temp_hc_actual(self) -> 'Series[float]':
        """Hourly air free temperature with heating/cooling adjustments set to actual requirements in W/m2.
        Ѳair,ac [Hourly Simulation column CT]
        """
        return self._hours['air_free_temp_hc_actual']

    @property
    def theta_m(self) -> 'Series[float]':
        """Hourly building thermal mass in degrees C.
        Ѳm [Hourly Simulation column BI]
        """
        return self._hours['theta_m']

    @property
    def rse(self) -> 'Series[float]':
        """Hourly external surface thermal resistance in m2K/W.
        [Hourly simulation column KU]  ISO 6946
        """
        if 'rse' not in self._hours.columns:
            conditions = [
                self.wind_speed <= 1.5,
                (self.wind_speed > 1.5) & (self.wind_speed <= 2.5),
                (self.wind_speed > 2.5) & (self.wind_speed <= 3.5),
                (self.wind_speed > 3.5) & (self.wind_speed <= 4.5),
                (self.wind_speed > 4.5) & (self.wind_speed <= 6.0),
                (self.wind_speed > 6.0) & (self.wind_speed <= 8.5),
                self.wind_speed > 8.0,
                ]
            choices = [
                0.08,
                0.06,
                0.05,
                0.04,
                0.04,
                0.03,
                0.02,
            ]
            self._hours['rse'] = select(conditions, choices, default=0.02)
        return self._hours['rse']

    @property
    def solar_heat_windows(self) -> 'Series[float]':
        """Hourly solar heat gains through windows in W/m2.
        ϕsol,w [Hourly Simulation column LH]

        Wattage is given by the sum of solar radiation on each window multiplied by its
        area and solar heat gain coefficient.
        Solar radiation is a function of climate data and building orientation.
        """
        return self._hours['solar_heat_windows']

    def get_solar_heat_opaque(self, row: Series) -> 'Series[float]':
        """Calculate solar heat gains through opaque surfaces for a specific orientation.
        [Hourly Simulation columns LK:LR]
        """
        compass_point = row.name
        solar_radiation = self.solar_irradiation.get_solar_irradiation(compass_point)  # M:T

        absorption = self.spec.parameters.facade_absorption_coefficient  # LK78:LR78
        view_factor = self.spec.parameters.view_factor_to_sky_facade  # LK80:LR80
        hr = 5 * self.spec.parameters.facade_emissivity  # [LK81:LR81 via Inputs E224]
        delta_theta_er = 11  # [Hardcoded in LK82: ISO 13790, 11.4.6]
        Fsh_ob_overhand = 1.0  # [Hardcoded in LK87:LR87 via Inputs C108:J108 = 0]
        # LR90 via Inputs C115:J115
        Fsh_ob_own = (
                (self.spec.building_length + self.spec.building_width) /
                (self.geometry.equivalent_rectangle.length + self.geometry.equivalent_rectangle.width)
        )
        Fsh_ob = Fsh_ob_overhand * Fsh_ob_own # LK95:LR95
        u_value = self.spec.uvalue_facade  # LK104:LR104
        opaque_area = row['opaque_area']  # LK105:LR105
        Rse = self.rse  # KU: ISO 6946
        return maximum(
            Fsh_ob *
            (absorption * Rse * opaque_area * u_value) *
            solar_radiation - (
                    view_factor * (Rse * u_value * opaque_area * hr * delta_theta_er)
            ),
            0.0
        )

    @property
    def solar_heat_roof(self):
        """Calculate solar heat gains through roof surfaces in W.
        HOR [Hourly Simulation column LS]
        """
        absorption = self.spec.parameters.roof_absorption_coefficient  # LS78
        view_factor = self.spec.parameters.view_factor_to_sky_roof  # LS80
        hr = 5 * self.spec.parameters.facade_emissivity  # [LS81 via Inputs E224]
        delta_theta_er = 11  # [Hardcoded in LK82: ISO 13790, 11.4.6]
        Fsh_ob = 1  # [Hardcoded in LS95 ISO 13790, 11.4.4]
        u_value = self.spec.uvalue_roof # LS104
        area = self.geometry.roof_projections.sum()  # LS105
        rse = self.rse  # KU
        irradiation = self.solar_irradiation.ghi  # Hourly simulation column U
        return (
                Fsh_ob * (absorption * rse * area * u_value) * irradiation -
                (view_factor * (rse * u_value * area * hr * delta_theta_er))
        ).apply(lambda x: max(0, x))

    @property
    def solar_heat_opaque(self) -> 'Series[float]':
        """Hourly solar heat gains through opaque surfaces in W/m2.
        ϕsol,op [Hourly Simulation column LT]

        Wattage is given by the sum of solar radiation on each opaque surface multiplied by its
        area and solar heat gain coefficient.
        Solar radiation is a function of climate data and building orientation.
        Horizontal solar radiation is also included because of roof surfaces.
        """
        if 'solar_heat_opaque' not in self._hours.columns:
            opaque_facade_by_orientation = (
                self.geometry.opaque_areas
                .to_frame('opaque_area')
                .groupby('compass_point')
                .sum()
                .apply(
                    self.get_solar_heat_opaque,
                    axis=1
                )
            )
            self._hours['solar_heat_opaque'] = opaque_facade_by_orientation.sum() + self.solar_heat_roof
        return self._hours['solar_heat_opaque']

    @property
    def solar_heat(self) -> 'Series[float]':
        """Hourly solar heat gains in W/m2.
        ϕsol [Hourly Simulation column AJ, KO]

        Wattage per square meter is given by the sum of solar heat gains through windows
        and opaque surfaces, divided by the conditioned floor area.
        """
        return self._hours['solar_heat']

    @property
    def heating_cooling_demand(self) -> 'Series[float]':
        """Hourly temperature change demand based in W.
        ϕHC,nd W [Hourly simulation column DB * zone area]

        Has columns for heating and cooling demand because they need to be weighted by their scaling factors.

        This is used for the ASHRAE140 test cases to determine the heating and cooling demand.
        """
        return self._hours['heating_cooling_demand']

    @property
    def heating_demand(self) -> 'Series[float]':
        """Hourly heating need in W/m2.
        Represents values from heating_cooling_demand clipped to only positive values.
        """
        if 'heating_need' not in self._hours.columns:
            self._hours['heating_need'] = self.heating_cooling_demand.clip(0, None)
        return self._hours['heating_need']

    @property
    def cooling_demand(self) -> 'Series[float]':
        """Hourly cooling need in W/m2.
        Represents values from heating_cooling_demand clipped to only negative values, then inverted.
        """
        if 'cooling_need' not in self._hours.columns:
            self._hours['cooling_need'] = self.heating_cooling_demand.clip(None, 0)
            self._hours['cooling_need'] = -self._hours['cooling_need']
        return self._hours['cooling_need']


import logging
from pandas import DataFrame

from .base import EnergyUseSimulation
from .occupancy import OccupationSimulation
from ..types import OpenBESSpecification

logger = logging.getLogger(__name__)


class HotWaterSimulation(EnergyUseSimulation):
    def __init__(self, spec: OpenBESSpecification, occupancy: OccupationSimulation = None):
        super().__init__(spec=spec)
        self.occupancy = occupancy or OccupationSimulation(spec=spec)

    def get_daily_hot_water_nominal(self) -> float:
        """Calculate nominal (pre-efficiency scaling) daily hot water energy consumption.
        Args:
            spec (OpenBESSpecification): The building specifications spec data class.
        Returns:
            float: Nominal daily hot water energy consumption in kWh.
        """
        demand = self.spec.water_demand  # l/day
        specific_heat_capacity_water = 4.18  # J/g°C
        output_temp = self.spec.water_reference_temperature  # °C
        input_temp = self.spec.water_supply_temperature  # °C
        per_hour = 1 / 3_600  # Convert seconds to hours

        if demand is None or output_temp is None or input_temp is None:
            logger.warning("Insufficient data to calculate hot water energy consumption.")
            return 0.0

        temperature_rise = output_temp - input_temp  # °C (from cold to hot water)

        return specific_heat_capacity_water * temperature_rise * demand * per_hour

    def get_daily_hot_water(self) -> float:
        """
        Adjust nominal hot water energy consumption by heater efficiency.
        Args:
            spec (OpenBESSpecification): The building specifications spec data class.
        Returns:
            float: Daily hot water energy consumption in kWh.
        """
        if self.spec.water_system_efficiency_cop is None:
            logger.warning("Insufficient data to calculate hot water energy consumption.")
            return 0.0

        return self.get_daily_hot_water_nominal() * self.spec.water_system_efficiency_cop

    def get_hot_water_per_month(self) -> DataFrame:
        """Return the amount of energy used heating water for each month of the year.
        Args:
            spec (OpenBESSpecification): The building specifications spec data class.
        Returns:
            DataFrame: Hot water energy consumption in kWh for each month.
        """
        kWh_per_day = self.get_daily_hot_water()
        result = self.occupancy.occupied_days_per_month * kWh_per_day
        result.name = "kWh"
        return result

    @property
    def energy_use(self) -> DataFrame:
        """Hot water energy use in kWh for each hour of the year.
        """
        if self._energy_use[self.spec.water_system_energy_source].hasnans:
            daily_kWh = self.get_daily_hot_water()
            hourly_kWh = (
                    (daily_kWh * self.occupancy.occupied_days_per_month.sum()) /
                    self.occupancy.occupancy['is_occupied'].sum()
            )
            self._energy_use[self.spec.water_system_energy_source] = (
                    [hourly_kWh] * self.occupancy.occupancy['is_occupied']
            )
        return self._energy_use